# -*- coding: utf-8 -*-

from .fits_kw import FITS_KEYWORDS, getFitsKw
from ..dataset.arraydataset import ArrayDataset
from ..dataset.tabledataset import TableDataset
from ..dataset.dataset import CompositeDataset
from ..dataset.unstructureddataset import UnstructuredDataset
from ..dataset.dataset import Dataset
from ..dataset.datatypes import DataTypes
from ..dataset.baseproduct import BaseProduct
from ..dataset.dateparameter import DateParameter
from ..dataset.stringparameter import StringParameter
from ..dataset.numericparameter import NumericParameter, BooleanParameter
from ..dataset.datatypes import Vector

import os
from collections.abc import Sequence
import io

FITS_INSTALLED = True
try:
    import numpy as np
    from astropy.io import fits
    from astropy.table import Table
    from astropy.table import Column
except ImportError:
    FITS_INSTALLED = False

debug = False
typecode2np = {
    "b": np.int8,    # signed char
    "B": np.uint8,   # unsigned char
    "u": str,     # string
    "h": np.int16,   # signed short
    "H": np.uint16,  # unsigned integer
    "i": np.int16,   # signed integer
    "I": np.uint16,  # unsigned integer
    "l": np.int32,   # signed long
    "L": np.uint32,  # unsigned long
    "q": np.int64,   # signed long long
    "Q": np.uint64,  # unsigned long long
    "f": np.float32,  # float
    "d": np.float64,   # double
    "c": np.complex64,  # complex
    "c128": np.complex128,  # complex 128 b
    "t": bool,       # truth value
    "V": np.void,       # raw bytes block of fixed length
    "U": str
}


def main():
    fitsdir = '/Users/jia/desktop/vtse_out/'
    if os.path.exists(fitsdir + 'array.fits'):
        os.remove(fitsdir + 'array.fits')
    ima = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8]], description='a')
    imb = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8], [
                       1, 2, 3, 4], [5, 6, 7, 8]], description='b')
    # im=[[1,2,3,4],[5,6,7,8]]
    hdul = fits.HDUList()
    fits_dataset(hdul, [ima, imb])


def toFits(data, file='', **kwds):
    """convert dataset to FITS.

    :data: a list of Dataset or a BaseProduct (or its subclass).
    :file: '' for returning fits stream. string for file name. default ''.
    """
    # __import__('pdb').set_trace()
    hdul = fits.HDUList()
    hdul.append(fits.PrimaryHDU())
    if issubclass(data.__class__, (BaseProduct)):
        sets = data.values()
        names = list(data.keys())
        hdul = fits_dataset(hdul, sets, names)
        add_header(data.meta, hdul[0].header)
        hdul[0].header['EXTNAME'] = 'PrimaryHDU'
    elif issubclass(data.__class__, (ArrayDataset, TableDataset, CompositeDataset)):
        if issubclass(data.__class__, (ArrayDataset)):
            # dataset -> fits
            sets = [data]
            names = ['IMAGE']
        elif issubclass(data.__class__, (TableDataset)):
            sets = [data]
            names = ['TABLE']
        elif issubclass(data.__class__, (CompositeDataset)):
            sets = data.values()
            names = list(data.keys())
        hdul = fits_dataset(hdul, sets, names)
        # when passed a dataset instead of a list, meta go to PrimaryHDU
        add_header(data.meta, hdul[0].header)
    elif issubclass(data.__class__, Sequence) and \
            issubclass(data[0].__class__, (ArrayDataset, TableDataset, CompositeDataset)):
        hdul = fits_dataset(hdul, data)
    else:
        raise TypeError(
            'Making FITS needs a dataset or a product, or a Sequence of them.')
    if file:
        hdul.writeto(file, **kwds)
        return hdul
    elif file == '':
        with io.BytesIO() as iob:
            hdul.writeto(iob, **kwds)
            fits_im = iob.getvalue()
        return fits_im
    else:
        return hdul


def fits_dataset(hdul, dataset_list, name_list=None, level=0):
    """ Fill an HDU list with dataset data.

    :hdul: `list` of HDUs. 
    :dataset_list: `list` of dataset subclasses.
    :name_list:
    """
    if name_list is None:
        name_list = []

    dataset_only = issubclass(
        dataset_list.__class__, (ArrayDataset, TableDataset, CompositeDataset))

    for n, ima in enumerate(dataset_list):
        header = fits.Header()
        if issubclass(ima.__class__, ArrayDataset):
            a = np.array(ima)
            if not dataset_only:
                header = add_header(ima.meta, header)
            ename = ima.__class__.__name__ if len(
                name_list) == 0 else name_list[n]
            header['EXTNAME'] = ename
            hdul.append(fits.ImageHDU(a, header=header))
        elif issubclass(ima.__class__, TableDataset):
            t = Table()
            for name, col in ima.items():
                tname = typecode2np['u' if col.typecode == 'UNKNOWN' else
                                    'u' if col.typecode.endswith('B') else
                                    col.typecode]
                if debug:
                    print('tname:', tname)
                c = Column(data=col.data, name=name, dtype=tname, shape=[
                ], length=0, description=col.description, unit=col.unit, format=None, meta=None, copy=False, copy_indices=True)
                t.add_column(c)
            if not dataset_only:
                header = add_header(ima.meta, header)
            ename = ima.__class__.__name__ if len(
                name_list) == 0 else name_list[n]
            header['EXTNAME'] = ename
            hdul.append(fits.BinTableHDU(t, header=header))
        elif issubclass(ima.__class__, CompositeDataset):
            if not dataset_only:
                header = add_header(ima.meta, header)
            hdul.append(fits.BinTableHDU(Table(), header=header))
            for name, dlist in ima.items():
                #print('dlist', dlist.__class__)
                fits_dataset(hdul, [dlist], name_list=[name], level=level+1)
        elif issubclass(ima.__class__, UnstructuredDataset):
            raise NotImplemented("UnstructuredDataset not yet supported")
        else:
            raise TypeError('Must be a Dataset to convert to fits.')
    if debug:
        print("****", len(hdul))
    return hdul

    #hdul.writeto(fitsdir + 'array.fits')

   # f = fits.open(fitsdir + 'array.fits')
   # print(len(f))
    #h1 = f[0].header
    #h2 = f[1].header
    # print(h2)
    # return h1


def add_header(meta, header):
    """ Populate  header with keyword lines extracted from MetaData.

    :meta: :class: `MetaData`
    """
    for name, param in meta.items():
        pval = param.value
        if pval is None:
            v = fits.card.Undefined()
            kw = getFitsKw(name)
            header[kw] = (v, param.description)
        elif issubclass(param.__class__, DateParameter):
            value = pval.isoutc() if pval.tai else fits.card.Undefined()
            kw = getFitsKw(name)
            header[kw] = (value, param.description)
        elif issubclass(param.__class__, NumericParameter):
            if issubclass(pval.__class__, (Sequence)):
                for i, com in enumerate(pval):
                    kw = getFitsKw(name, ndigits=1)+str(i)
                    header[kw] = (com, param.description+str(i))
                    if debug:
                        print(kw, com)
            else:
                kw = getFitsKw(name)
                header[kw] = (pval, param.description)
        elif issubclass(param.__class__, StringParameter):
            kw = getFitsKw(name)
            if pval == 'UNKNOWN':
                v = fits.card.Undefined()
            else:
                v = pval
            header[kw] = (v, param.description)
        elif issubclass(param.__class__, BooleanParameter):
            kw = getFitsKw(name)
            v = 'T' if pval else 'F'
            header[kw] = (v, param.description)
        else:
            kw = getFitsKw(name)
            v = fits.card.Undefined()
            header[kw] = (v, '%s of unknown type' % str(pval))
    if debug:
        print('*** add_header ', header)
    return header


def fits_header():
    fitsdir = '/Users/jia/desktop/vtse_out/'
    f = fits.open(fitsdir + 'array.fits')
    h = f[0].header
    #h.set('add','header','add a header')
    h['add'] = ('header', 'add a header')
    h['test'] = ('123', 'des')
    f.close()

    return h


def test_fits_kw(h):
    # print(h)
    # print(list(h.keys()))
    #assert FITS_KEYWORDS['CUNIT'] == 'cunit'
    assert getFitsKw(list(h.keys())[0]) == 'SIMPLE'
    assert getFitsKw(list(h.keys())[3]) == 'NAXIS1'


if __name__ == '__main__':

    # test_fits_kw(fits_data())
    main()



# -*- coding: utf-8 -*-

from . import metadata

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class MetaDataHolder():
    """ Object holding meta data. 

    """

    def __init__(self, meta=None, **kwds):
        """ Adds MetaData to the class.
        with defaults set to self.zInfo['metadata'].

        Parameters
        ----------

        Returns
        -------
        """
        if meta is None:
            meta = metadata.MetaData()
        self.setMeta(meta)
        super().__init__(**kwds)

    def getMeta(self):
        """ Returns the current MetaData container of this object. 
        Cannot become a python property because setMeta is in Attributable

        Parameters
        ----------

        Returns
        -------

        """
        return self._meta

    def hasMeta(self):
        """ whether the metadata holder is present.
        During initialization subclass of MetaDataHolder may need to know if the metadata holder has been put in place with is method.

        Parameters
        ----------

        Returns
        -------

        """
        return hasattr(self, '_meta')

    def setMeta(self, meta):
        """
        Parameters
        ----------

        Returns
        -------

        """
        self._meta = meta



# -*- coding: utf-8 -*-

from .serializable import Serializable

import logging
from collections.abc import Mapping, Sequence, Set
from itertools import chain
from collections import OrderedDict
import array
import decimal
import datetime
import fractions
import sys
import hashlib

HASH_WIDTH = sys.hash_info.width // 8

if sys.version_info[0] + 0.1 * sys.version_info[1] >= 3.6:
    PY36 = True
else:
    PY36 = False

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class CircularCallError(RuntimeError):
    pass


DEEPCMP_RESULT = None


def deepcmp(obj1, obj2, seenlist=None, verbose=False, eqcmp=False, brief=True):
    """ Recursively descends into obj1's every component and
    compares with its counterpart in obj2.

    Ways to test includes (in this order):

    * if they are the same object
    * type
    * ```__eq__``` or ```__cmp__``` if requested
    * state from ```__getstate__()```
    * quick length
    * members if is ```Mapping```, ```Sequence``` (
    tuple, set, list, dict, OrderedDict, UserDict ... )
    * properties/attributes in ```__dict__```
    * ```__eq__``` or ```__cmp__``` even if not requested

    Detects cyclic references.

    Returns
    -------
    ``None`` if finds no difference, ```1``` if `brief is ```True``` else a string of explanation.

    :eqcmp: if True, use __eq__ or __cmp__ if the objs have them. If False only use as the last resort. default True.
    """

    global DEEPCMP_RESULT

    # seen and level are to be used as nonlocal variables in run()
    # to overcome python2's lack of nonlocal type this method is usded
    # https://stackoverflow.com/a/28433571
    class _context:
        if seenlist is None:
            seen = []
        else:
            seen = seenlist
        level = 0

    def run(o1, o2, explain=False, v=False, eqcmp=True, brief=True):
        """
        Paremeters
        ----------
        :o1: object to be compared.
        :o2: object to be compared.
        :v: Prints out step-by-step comparison if set `True`.
        :brief: if not equal, return 1 if set `True` else returns an explaination.
        :eqcmp: use ```__eq__``` and ```__cmp__``` for comparison.

        Returns
        -------
        ```None``` if equal. ```1``` if ```brief``` is ```True``` else a string of explanation why not equal.

        """
        #
        # nonlocal seen
        # nonlocal level
        if v or not brief:
            from ..utils.common import lls, bstr
        id1, id2 = id(o1), id(o2)
        if id1 == id2:
            if v:
                print('These are the same object o1=%s ||| o2=%s.' %
                      (bstr(o1, 20), bstr(o2, 20)))
            return None
        pair = (id1, id2) if id1 < id2 else (id2, id1)
        c = o1.__class__
        c2 = o2.__class__

        _context.level += 1
        if v:
            print('deepcmp level %d seenlist length %d' %
                  (_context.level, len(_context.seen)))
            print('1 ' + str(c) + lls(o1, 45))
            print('2 ' + str(c2) + lls(o2, 45))
        if pair in _context.seen:
            msg = 'deja vue %s' % str(pair)
            raise CircularCallError(msg)
        _context.seen.append(pair)
        if c != c2:
            if v:
                print('type diff')
            _context.level -= 1
            del _context.seen[-1]
            return 1 if brief else ' due to diff types: ' + c.__name__ + ' and ' + c2.__name__
        if issubclass(c, (int, float, complex, str, bytes, bool, datetime.datetime)):
            if v:
                print('find simple type')
            _context.level -= 1
            del _context.seen[-1]
            if o1 != o2:
                return 1 if brief else ' due to difference: "%s" ||| "%s"' % (o1, o2)
            else:
                return None

        has_eqcmp = (hasattr(o1, '__eq__') or hasattr(
            o1, '__cmp__')) and not issubclass(c, DeepEqual)
        if eqcmp and has_eqcmp:
            if v:
                print('obj1 has __eq__ or __cmp__ and not using deepcmp')
            # checked in-seen to ensure whst follows will not cause RecursionError
            try:
                t = o1 == o2
            except CircularCallError as e:
                if v:
                    print('Get circular call using eq/cmp: '+str(e))
                pass
            else:
                _context.level -= 1
                del _context.seen[-1]
                if t:
                    return None
                else:  # o1 != o2:
                    return 1 if brief else \
                        ' due to "%s" != "%s"' % (lls(o1, 155), lls(o2, 155))

        if hasattr(o1, '__getstate__'):
            if v:
                print('Find __getstate__')
            try:
                o1 = o1.__getstate__()
                o2 = o2.__getstate__()
            except TypeError:
                logger.error('__getstate__ trouble')
                raise
            else:  # no exception for __getstate__
                r = run(o1, o2, v=v, eqcmp=eqcmp, brief=brief)
                del _context.seen[-1]
                _context.level -= 1
                if r:
                    return 1 if brief else ' due to o1.__getstate__ != o2.__getstate__' + r
                else:
                    return None

        try:
            # this is not good if len() is delegated
            # if hasattr(o1, '__len__') and len(o1) != len(o2):
            if hasattr(o1, '__len__') and len(o1) != len(o2):
                del _context.seen[-1]
                _context.level -= 1
                return 1 if brief else \
                    ' due to diff %s lengths: %d and %d (%s, %s)' %\
                    (c.__name__, len(o1), len(o2), lls(
                        list(o1), 115), lls(list(o2), 115))
        except AttributeError:
            pass

        if issubclass(c, Mapping):
            if v:
                print('Find Mapping')
                print('check keys')

            from .odict import ODict
            if issubclass(c, (OrderedDict, ODict)) or PY36:
                #
                r = run(list(o1.keys()), list(o2.keys()),
                        v=v, eqcmp=eqcmp, brief=brief)
            else:
                #  old dict or UserDict
                r = run(tuple(sorted(o1.keys(), key=hash)),
                        tuple(sorted(o1.keys(), key=hash)),
                        v=v, eqcmp=eqcmp, brief=brief)
            if r is not None:
                del _context.seen[-1]
                _context.level -= 1
                return 1 if brief else " due to diff " + c.__name__ + " keys" + r
            if v:
                print('check values')
            for k in o1.keys():
                if k not in o2:
                    del _context.seen[-1]
                    _context.level -= 1
                    return 1 if brief else ' due to o2 has no key=%s' % (lls(k, 155))
                r = run(o1[k], o2[k], v=v, eqcmp=eqcmp, brief=brief)
                if r is not None:
                    del _context.seen[-1]
                    _context.level -= 1
                    return 1 if brief else ' due to diff values for key=%s' % (lls(k, 155)) + r
            del _context.seen[-1]
            _context.level -= 1
            return None
        elif issubclass(c, (Set, Sequence)):
            if v:
                print('Find Set, Sequence.')
            if issubclass(c, Sequence):
                if v:
                    print('Check Sequence.')
                for i in range(len(o1)):
                    r = run(o1[i], o2[i], v=v, eqcmp=eqcmp, brief=brief)
                    if r is not None:
                        del _context.seen[-1]
                        _context.level -= 1
                        return 1 if brief else ' due to diff at index=%d (%s %s)' % \
                            (i,
                             lls(o1[i], 10),
                             lls(o2[i], 10)) + r
                _context.level -= 1
                del _context.seen[-1]
                return None
            else:
                if v:
                    print('Check Set.')
                if 1:
                    del _context.seen[-1]
                    _context.level -= 1
                    if o1.difference(o2):
                        return 1 if brief else ' due to at least one element in o1 not in o2'
                    else:
                        return None
                else:
                    oc = o2.copy()
                    for m in o1:
                        found = False
                        for n in oc:
                            r = run(m, n, v=v, eqcmp=eqcmp, brief=brief)
                            if r is None:
                                found = True
                                break
                        if not found:
                            del _context.seen[-1]
                            _context.level -= 1
                            return 1 if brief else ' due to %s not in the latter' % (lls(m, 155))
                        oc.remove(n)
                    del _context.seen[-1]
                    _context.level -= 1
                    return None
        else:
            if hasattr(o1, '__dict__'):
                if v:
                    print('obj1 has __dict__')
                    o1 = sorted(vars(o1).items())
                    o2 = sorted(vars(o2).items())
                    r = run(o1, o2, v=v, eqcmp=eqcmp, brief=brief)
                    del _context.seen[-1]
                    _context.level -= 1
                if r:
                    return 1 if brief else ' due to o1.__dict__ != o2.__dict__' + r
                else:
                    return None
            # elif hasattr(o1, '__iter__') and hasattr(o1, '__next__') or \
            #         hasattr(o1, '__getitem__'):
            #     # two iterators are equal if all comparable properties are equal.
            #     del _context.seen[-1]
            #     _context.level -= 1
            #     return None
            elif has_eqcmp:
                # last resort
                if o1 == o2:
                    del _context.seen[-1]
                    _context.level -= 1
                    return None
                else:
                    del _context.seen[-1]
                    _context.level -= 1
                    return 1 if brief else ' according to __eq__ or __cmp__'
            else:  # o1 != o2:
                del _context.seen[-1]
                _context.level -= 1
                if v:
                    print('no way')
                return 1 if brief else ' due to no reason found for "%s" == "%s"' % \
                    (lls(o1, 155), lls(o2, 155))
    res = run(obj1, obj2, v=verbose, eqcmp=eqcmp, brief=brief)
    DEEPCMP_RESULT = res
    return res


XHASH_VERBOSE = False


def xhash(hash_list=None, seenlist=None, verbose=None):
    """ get the hash of a tuple of hashes of all members of given sequence.

    :hash_list: use instead of self.getstate__()
    :verbose: set to trace.
    """

    if verbose is None:
        verbose = XHASH_VERBOSE

    # https://stackoverflow.com/a/28433571
    class _context:
        if seenlist is None:
            seen = []
        else:
            seen = seenlist
        level = 0

    def run(hash_list=None):
        _context.level += 1
        ind = ' ' * _context.level
        hashes = []

        if verbose:
            from ..utils.common import lls, bstr

        if 0 and verbose:
            print('entering id%d id%d lv%d len%d' % (id(_context.level), id(_context.seen),
                                                     _context.level, len(_context.seen)))
        hlid = id(hash_list)
        if hlid in _context.seen:
            if verbose:
                print(ind + 'seen it')
            _context.level -= 1
            del _context.seen[-1]
            return 0
        _context.seen.append(hlid)
        if issubclass(hash_list.__class__, int):
            res = hash_list
            if verbose:
                print(ind + 'int "%s" -- %s' % (lls(hash_list, 20), res))
            _context.level -= 1
            del _context.seen[-1]
            return res
        elif issubclass(hash_list.__class__, (float, decimal.Decimal, fractions.Fraction)):
            res = hash(hash_list)
            if verbose:
                print(ind + '%s "%s" -- %s' %
                      (hash_list.__class__.__name__, lls(hash_list, 20), res))
            _context.level -= 1
            del _context.seen[-1]
            return res
        elif issubclass(hash_list.__class__, (str, bytes)):
            # put str first so it is not treated as a sequence
            res = hash(hash_list)
            if verbose:
                print(ind + 'str/bytes "%s" -- %s' %
                      (lls(hash_list, 20), res))
            _context.level -= 1
            del _context.seen[-1]
            return res
        elif issubclass(hash_list.__class__, (array.array)):
            hasher = hashlib.new('sha256', hash_list.typecode.encode('utf-8'))
            hasher.update(hash_list)
            res = int.from_bytes(
                hasher.digest()[:HASH_WIDTH], byteorder=sys.byteorder)
            # source = (hash_list.typecode,
            #           hash_list.itemsize,
            #           len(hash_list),
            #           len(hash_list[0]))
            if verbose:
                print(ind + '%s %s %s' % (hash_list.__class__.__name__,
                                          lls(hash_list, 20), res))
            _context.level -= 1
            del _context.seen[-1]
            return res
        elif hasattr(hash_list, '__getstate__'):
            try:
                o = hash_list.__getstate__()
            except TypeError:
                logger.error('__getstate__ trouble')
                raise
            else:  # no exception for __getstate__
                source = chain.from_iterable(o.items())
            if verbose:
                print(ind + '%s %s %s' % (hash_list.__class__.__name__,
                                          lls(source, 20), 'has __getstate__'))
        elif issubclass(hash_list.__class__, (Set, Sequence)):
            source = hash_list
            if verbose:
                print(ind + '%s %s %s' % (hash_list.__class__.__name__,
                                          lls(source, 20), 'is Sequence'))
        elif issubclass(hash_list.__class__, Mapping):
            source = chain.from_iterable(hash_list.items())
            if verbose:
                print(ind + '%s %s %s' % (hash_list.__class__.__name__,
                                          lls(source, 20), 'is Mapping'))
        else:
            res = hash(hash_list)
            if verbose:
                print(ind + '%s %s -- %s' % (hash_list.__class__.__name__,
                                             lls(hash_list, 20), res))
            _context.level -= 1
            del _context.seen[-1]
            return res

        for t in source:
            if hasattr(t, 'hash'):
                h = t.hash()
            else:
                h = run(t)
            if verbose:
                print(ind + '> %s %s -- %s' % (h.__class__.__name__,
                                               lls(h, 20), h))
            hashes.append(h)
        # if there is only one element only hash the element
        res = hash(hashes[0] if len(hashes) == 1 else tuple(hashes))
        if verbose:
            print(ind + '%s %s -- %s' % ('RET', str(len(hashes)), res))
        _context.level -= 1
        del _context.seen[-1]
        return res
    return run(hash_list=hash_list)


class DeepcmpEqual(object):
    """ mh: Can compare key-val pairs of another object
    with self.

    False if compare with None
    or exceptions raised, e.g. objects that do not have items()
    """

    def __init__(self, **kwds):
        super().__init__(**kwds)

    def equals(self, obj, verbose=False, brief=True):
        """
        Paremeters
        ----------

        :verbose: Prints out step-by-step comparison if set `True`.

        Returns
        -------

        """
        r = self.diff(obj, [], verbose=verbose, brief=brief)
        # logging.debug(r)
        return r is None

    def __eq__(self, obj):
        """
        Paremeters
        ----------

        Returns
        -------

        """
        return self.equals(obj)

    def __ne__(self, obj):
        """
        Paremeters
        ----------

        Returns
        -------

        """

        return not self.__eq__(obj)

    def diff(self, obj, seenlist, verbose=False, brief=True):
        """ recursively compare components of list and dict
        until finding equality or exhausting all ways of testing.

        :seenlist: a list of classes that has been seen. will not descend in to them.
        :verbose: Prints out step-by-step comparison if set `True`.
        :brief: if not equal, return 1 if set `True` else returns an explaination.
        Paremeters
        ----------

        Returns
        -------
        """
        if issubclass(self.__class__, Serializable):
            if issubclass(obj.__class__, Serializable):
                r = deepcmp(self.__getstate__(),
                            obj.__getstate__(),
                            seenlist=seenlist,
                            verbose=verbose,
                            brief=brief)
            else:
                return 1 if brief else ('different classes')
        else:
            r = deepcmp(self, obj, seenlist=seenlist,
                        verbose=verbose, brief=brief)
        return r


class EqualDict(object):
    """ mh: Can compare key-val pairs of another object
    with self.

    False if compare with None
    or exceptions raised, e.g. obj  that do not have items().
    """

    def __init__(self, **kwds):
        super().__init__(**kwds)

    def equals(self, obj, verbose=False):
        """
        Paremeters
        ----------

        Returns
        -------

        """

        if obj is None:
            return False
        try:
            if self.__dict__ != obj.__dict__:
                if verbose:
                    print('@@ diff ' + lls(self.__dict__) +
                          '\n>>diff \n' + lls(obj.__dict__))
                return False
        except Exception as err:
            # print('Exception in dict eq comparison ' + lls(err))
            return False
        return True

    def __eq__(self, obj):
        """
        Paremeters
        ----------

        Returns
        -------

        """

        return self.equals(obj)

    def __ne__(self, obj):
        """
        Paremeters
        ----------

        Returns
        -------

        """

        return not self.__eq__(obj)


class EqualODict(object):
    """ mh: Can compare order and key-val pairs of another object
    with self. False if compare with None
    or exceptions raised, e.g. obj that do not have items()
    """

    def __init__(self, **kwds):
        super().__init__(**kwds)

    def equals(self, obj, verbose=False):
        """
        Paremeters
        ----------

        Returns
        -------

        """
        if obj is None:
            return False
        try:
            return list(self.items()) == list(obj.items())
        except Exception:
            return False
        return True

    def __eq__(self, obj):
        """
        Paremeters
        ----------

        Returns
        -------

        """
        return self.equals(obj)

    def __ne__(self, obj):
        """
        Paremeters
        ----------

        Returns
        -------

        """

        return not self.__eq__(obj)


class StateEqual():
    """ Equality tested by hashed state.
    """

    def __init__(self, *args, **kwds):
        """ Must pass *args* so `DataWrapper` in `Composite` can get `data`.
        """

        super().__init__(*args, **kwds)  # StateEqual

    def hash(self, **kwds):
        return xhash(self.__getstate__(), **kwds)

    def __eq__(self, obj, **kwds):
        """ compares hash. """

        if obj is None:
            return False

        if id(self) == id(obj):
            return True

        if type(self) != type(obj):
            return False
        try:
            h1 = self.hash()
            h2 = obj.hash()
        except AttributeError:
            return False
        # print('hashes ', h1, h2)
        return h1 == h2

    equals = __eq__

    def __xne__(self, obj):
        return not self.__eq__(obj)

    def __hash__(self, **kwds):
        return self.hash(**kwds)

    __hash__ = hash


class DeepcmpEqual():
    """ Equality tested by `deepcmp`.
    """

    def __init__(self, *args, **kwds):
        """ Must pass *args* so `DataWrapper` in `Composite` can get `data`.
        """

        super().__init__(*args, **kwds)  # DeepcmpEqual

    def __eq__(self, obj, **kwds):
        """ compares using `deepcmp`. """

        if obj is None:
            return False

        if id(self) == id(obj):
            return True

        if type(self) != type(obj):
            return False
        res = deepcmp(self, obj, **kwds)
        return res is None

    equals = __eq__

    def __xne__(self, obj):
        return not self.__eq__(obj)

    def __hash__(self, **kwds):
        return self.hash(**kwds)

    __hash__ = hash


DeepEqual = DeepcmpEqual



# -*- coding: utf-8 -*-

from .metadata import guess_value, Parameter
from .dataset import CompositeDataset
from .arraydataset import ArrayDataset, Column
from .tabledataset import TableDataset
from .serializable import Serializable

import itertools
from collections import OrderedDict, UserDict
import datetime
import array

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


def check_input(arg, serializable=True):
    """ Raise exception if needed when arg is not simple type or Serializable.
    """
    cls = arg.__class__
    nm = cls.__name__
    if issubclass(cls, (int, float, str, bool, bytes, complex,
                        list, dict, array.array, UserDict,
                        datetime.datetime, type(None))):
        return arg

    if serializable:
        if issubclass(cls, tuple) or not issubclass(cls, Serializable):
            raise TypeError(
                'History parameter "%s" not found to be serializable by `json.dump()` or a subclass of `Serializable`.' % nm)
    return arg


class History(TableDataset):
    """ Public interface to the history dataset. Contains the
    main methods for retrieving a script and copying the history.
    """

    def __init__(self, **kwds):
        """

        Implemented as a `TableDataset`. Argument and context information
        are stored as metadata key-variable pairs in the metadata.
        Input data are in the table where name is the column name and
        references in the first cell of columns.

        mh: The copy constructor is better not be implemented. Use copy()
        instead. Remember: not only copies the datasets,
        but also changes the history ID in the metadata and
        relevant table entries to indicate that this a new
        independent product of which the history may change.

        Parameters
        ----------

        Returns
        -------

        """
        super(History, self).__init__(**kwds)
        self['name'] = Column([], '')
        self['reference'] = Column([], '')
        self.description = 'Named positional and keyword arguments, relevant context, and input data to the pipeline or task that generated this product.'
        self.builtin_keys = list(self._meta.keys())

    def accept(self, visitor):
        """ Hook for adding functionality to meta data object
        through visitor pattern.

        Parameters
        ----------

        Returns
        -------

        """
        visitor.visit(self)

    def getScript(self):
        """ Creates a script from the history.

        Parameters
        ----------

        Returns
        -------

        """

        return self._meta['command_line'].value()

    def getTaskHistory(self, format='graph', node=None, use_name=True, verbose=False):
        """ Returns a human readable formatted history tree.

        Parameters
        ----------
        format : str
            Output format: `graph' (default) for `networkx.DiGraph`; 'ascii" for dump; 'png', 'svg', 'jpg' for graphic formats.
        node : str
            A name that uniquely identifies the parent product.

        Returns
        -------
        `networkx.DiGraph`:
            A graph of input data names and references.
        """
        from ..pal.urn import is_urn
        from ..pal.productref import ProductRef
        import networkx
        if node is None:
            node = 'root'
        new_g = networkx.DiGraph()
        h = None
        dt = self._data
        if verbose:
            print('History graph for %s has %d inputs: %s.' %
                  (node, len(dt['name']), str(list(dt['name']))))
        # __import__('pdb').set_trace()

        for name, ref in zip(dt['name'], dt['reference']):
            # pydot wants no unquoted :
            refq = f'"{ref}"'
            if use_name:
                ref_node = (name, {'product_reference': refq})
            else:
                ref_node = (refq, {'product_name': name})
            if verbose:
                print(f"Node {ref_node}")
            new_g.add_nodes_from([ref_node])
            new_g.add_edge(name if use_name else refq, node)
            if is_urn(ref):
                inp = ProductRef(ref).getProduct()
                if verbose:
                    print(f'Load product {inp.description}:{id(inp)}.')
                # get a graph with a node named with ref
                h = inp.history.getTaskHistory(
                    node=name if use_name else refq,
                    use_name=use_name, verbose=verbose)
                # Merge
                new_g = networkx.compose(new_g, h)
                if verbose:
                    print(
                        f'Returning history graph {h.adj} for {node}==>{new_g.adj}')
        if format == 'graph':
            return new_g
        if format == 'ascii':
            try:
                import pydot
            except ImportError:
                pass
            else:
                return networkx.drawing.nx_pydot.to_pydot(new_g).to_string()
            return str(new_g.succ)
        if format == 'aaasvg':
            pass
        if format in ('svg', 'png', 'jpg'):
            try:
                import pydot
            except ImportError:
                pass
            else:
                im = networkx.drawing.nx_pydot.to_pydot(
                    new_g).create(format=format)
                return im
            return newg_g

    def graph(self, format='png', **kwds):
        """ calling getTaskHistory with simplified parameters format, default to png."""

        return self.getTaskHistory(format=format, **kwds)

    def add_input(self, args=None, info=None, refs=None, **kwds):
        """Add an entry to History records.

        A general product history is made of a series of records, each added by a
        processing step, usually called a pipeline. The record can be
        added by this method to this `History` object..

        Parameters
        ----------
        args : dict
            A mapping of  argument names and their
            values. Can be `vars(ArgParse())`. Values must be serializablee.
        info : dict
            keywords and values in string.
        refs : dict
            A mapping of name and reference string pairs. The Reference string is the URN, file name, OSS address, URL, or other kind of pointer. The name is for human to identify the data, the reference for recursive retrieving inputs to the data.
        **kwds : dict
            appended to `info` by default.
        Returns
        -------
            result

        """
        if args or info or kwds:
            for name, var in itertools.chain(args.items(), info.items(), kwds.items()):
                cvar = check_input(var)
                # append the parameter name and value
                self._meta[name] = Parameter(value=cvar)

        if not refs:
            refs = {}
        for name, ref in refs.items():
            # append the name and input data reference
            if name in self.builtin_keys:
                name = name + '___'
            self._data['name']._data.append(name)
            self._data['reference']._data.append(ref)
            self._data['name'].updateShape()
            self._data['reference'].updateShape()
            self.updateShape()

    def get_args_info(self):
        """Get arguments and context information as a dictionary.

        Returns
        -------
        dict

        """

        return dict((n[:-3] if n.endswith('___') else n, v.value) for n, v in self._meta._data.items() if n not in self.builtin_keys)

    def xx__getstate__(self):
        """ Can be encoded with serializableEncoder

        Parameters
        ----------

        Returns
        -------

        """
        return OrderedDict(
            _ATTR_meta=self._meta,
            **self.data)



# -*- coding: utf-8 -*-

from fdi.utils.leapseconds import utc_to_tai, tai_to_utc, dTAI_UTC_from_utc, _fallback
from fdi.dataset.eq import deepcmp
from fdi.dataset.metadata import make_jsonable
from fdi.dataset.finetime import FineTime
from fdi.dataset.datatypes import Vector, Quaternion
from fdi.dataset.deserialize import Class_Look_Up, serialize_args, deserialize_args
from fdi.pal.urn import Urn
from fdi.pal.productref import ProductRef
from fdi.utils.checkjson import checkjson
from fdi.utils.loadfiles import loadcsv
from fdi.utils import moduleloader
from fdi.utils.common import fullname, wls, find_all_files
from fdi.utils.options import opt
from fdi.utils.fetch import fetch
from fdi.utils.tree import tree
from fdi.utils.loadfiles import loadMedia
from fdi.utils.getconfig import getConfig, Config_Look_Up, Config_NameSpace
import traceback
import importlib
import copy
from datetime import timezone, timedelta, datetime
import sys
import os
import getopt
import json
import hashlib
import os.path as op
import pytest
from astropy.io import fits

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

# make format output in /tmp/outputs.py
mk_outputs = 0
output_write = 'tests/outputs_utils.py'

if mk_outputs:
    with open(output_write, 'wt', encoding='utf-8') as f:
        f.write('# -*- coding: utf-8 -*-\n')

if __name__ == '__main__' and __package__ == 'tests':
    # run by python -m tests.test_dataset
    if not mk_outputs:
        from outputs_utils import out_tree
else:
    # run by pytest

    # This is to be able to test w/ or w/o installing the package
    # https://docs.python-guide.org/writing/structure/

    from pycontext import fdi

    if not mk_outputs:
        from outputs_utils import out_tree

    from logdict import logdict
    import logging
    import logging.config

    # create logger
    logging.config.dictConfig(logdict)
    logger = logging.getLogger()
    logger.debug('%s logging level %d' %
                 (__name__, logger.getEffectiveLevel()))

    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)


def test_get_demo_product(demo_product):
    v, related = demo_product
    assert v['Browse'].data[1:4] == b'PNG'
    # print(v.yaml())
    p = v.getDefault()
    assert p == v['measurements']
    aref = ProductRef(related)
    v.refs['a_reference'] = aref
    v.refs['a_different_name'] = aref
    r0 = v.refs
    p['dset'] = 'foo'
    # refs is always the last
    assert list(v.keys())[-1] == 'refs'
    assert r0 == v.refs
    # existing key
    p['dset'] = 'foo'
    assert list(v.keys())[-1] == 'refs'
    assert r0 == v.refs
    s = r0.string(level=0)
    # print(s)
    checkjson(v, dbg=0)
    checkgeneral(v)


def checkgeneral(v):
    # can always add attributes
    t = 'random'
    v.testattr = t
    assert v.testattr == t
    try:
        m = v.notexists
    except AttributeError as e:
        assert str(e).split()[-1] == "'notexists'", traceback.print_exc()
    except:
        traceback.print_exc()
        assert false


def chk_sample_pd(p):
    v, s = fetch(["description"], p)
    assert v == p.description
    assert s == '.description'
    # metadatax
    with pytest.raises(KeyError):
        e = p.meta['extra']

    v, s = fetch(["meta", "speed"], p)
    assert v == p.meta['speed']
    assert s == '.meta["speed"]'
    # parameter
    v, s = fetch(["meta", "speed", "unit"], p)
    assert v == 'meter'
    assert v == p.meta['speed'].unit
    assert s == '.meta["speed"].unit'

    v, s = fetch(["meta", "speed", "value"], p)
    assert v == Vector((1.1, 2.2, 3.3))
    assert v == p.meta['speed'].value
    assert s == '.meta["speed"].value'
    v, s = fetch(["meta", "speed", "valid"], p)
    mkj = make_jsonable({(1, 22): 'normal', (30, 33): 'fast'})
    assert v == mkj
    assert v == make_jsonable(p.meta['speed'].valid)
    assert s == '.meta["speed"].valid'
    # TODO written is string
    # [[[1, 22], 'normal'], [[30, 33], 'fast']]
    v, s = fetch(["meta", "speed", "valid", 0, 1], p)
    assert v == 'normal'
    assert v == p.meta['speed'].valid[0][1]
    assert s == '.meta["speed"].valid[0][1]'
    #
    # validate execution
    v, s = fetch(["meta", "speed", "isValid", ], p)
    assert v == True
    assert v == p.meta['speed'].isValid()
    assert s == '.meta["speed"].isValid()'

    # datasets
    #
    v, s = fetch(["Environment Temperature", "unit"], p)
    assert v == 'C'
    assert v == p['Environment Temperature'].unit
    assert s == '["Environment Temperature"].unit'

    v, s = fetch(["Environment Temperature", "data"], p)
    assert v == [768, 767, 766, 4.4, 4.5, 4.6, 5.4E3]
    assert v == p['Environment Temperature'].data
    assert s == '["Environment Temperature"].data'

    # dataset has a parameter
    v, s = fetch(["Environment Temperature", "T0", "tai"], p)
    assert v == FineTime('2020-02-02T20:20:20.0202').tai

    # a 2D array dataset in compositedataset 'measurements'
    v, s = fetch(["measurements", 'calibration', "unit"], p)
    assert v == 'count'
    assert v == p['measurements']['calibration'].unit
    assert s == '["measurements"]["calibration"].unit'

    # data of a column in tabledataset within compositedataset
    v, s = fetch(["measurements", "Time_Energy_Pos", "Energy", "data"], p)
    t = [x * 1.0 for x in range(len(v))]
    assert v == [2 * x + 100 for x in t]
    assert v == p['measurements']['Time_Energy_Pos']['Energy'].data
    assert s == '["measurements"]["Time_Energy_Pos"]["Energy"].data'
    ys, s = fetch(["measurements", "Time_Energy_Pos", "y"], p)
    zs, s = fetch(["measurements", "Time_Energy_Pos", "z"], p)
    # y^2 + z^2 = 100 for all t
    assert all((y*y + z*z - 100) < 1e-5 for y, z in zip(ys.data, zs.data))


def test_fetch(demo_product):

    # simple nested structure
    v = {1: 2, 3: 4}
    u, s = fetch([3], v)
    assert u == 4
    assert s == '[3]'
    v = {'1': 2, 3: 4}
    u, s = fetch(['1'], v)
    assert u == 2
    assert s == '["1"]'
    v.update(d={'k': 99})
    u, s = fetch(['d', 'k'], v)
    assert u == 99
    assert s == '["d"]["k"]'
    # 3 levels
    v = [4, v,  5]
    w = copy.deepcopy(v)
    u, s = fetch([0], w)
    assert u == 4
    u, s = fetch([2], w)
    assert u == 5
    assert s == '[2]'
    u, s = fetch([1, 'd', 'k'], w)
    assert u == 99
    assert s == '[1]["d"]["k"]'

    # path is str
    assert fetch([2], w) == fetch('2', w)
    assert fetch([1, 'd', 'k'], w) == fetch(
        '1/d/k', w) == fetch('1.d.k', w, sep='.')
    assert fetch('/1/d/k ', w) == fetch('1.d.k', w, sep='.')
    assert fetch(' ', w, re='asd') == (w, 'asd')

    # objects

    class al(list):
        ala = 0
        @ staticmethod
        def alb(): pass

        def alf(self, a, b=9):
            return a, b

        def __init__(self, *a, i=[8], **k):
            super().__init__(*a, **k)
            self.ald = i
        alc = {3: 4}
        ale = [99, 88]

    # init
    v = al(i=[1, 2])
    u, s = fetch(['ald'], v)
    assert u == [1, 2]
    assert s == '.ald'

    # class property list
    u, s = fetch(['ale', 1], v)
    assert u == 88
    assert s == '.ale[1]'

    # class property dict
    u, s = fetch(['alc', 3], v)
    assert u == 4
    assert s == '.alc[3]'

    # method
    u, s = fetch(['alb'], v)
    assert u is None
    assert s == '.alb()'

    # method w/ positional arg
    u, s = fetch(['alf__4.4'], v)
    assert u == (4.4, 9)
    assert s == '.alf(4.4)'

    # method w/ positional and keyword args
    allargs = serialize_args(4.4, [{"w": 77}, 65], not_quoted=True)
    assert allargs == '4.4__{"apiargs": [[{\"w\": 77}, 65]]}'
    u, s = fetch(['alf__' + allargs], v)
    assert u == (4.4, [{"w": 77}, 65])
    assert s == ".alf(4.4, [{'w': 77}, 65])"

    # method/function result
    u, s = fetch(['alf__' + allargs, 1, 0, 'w'], v)
    assert u == 77
    assert s == ".alf(4.4, [{'w': 77}, 65])[1][0][\"w\"]"

    class ad(dict):
        ada = 'p'
        adb = al([0, 6])

    v = ad(z=5, x=['b', 'n', {'m': 'j'}])
    v.ade = 'adee'

    u, s = fetch(['ada'], v)
    assert u == 'p'
    assert s == '.ada'

    u, s = fetch(['adb', 'ald', 0], v)
    assert u == 8
    assert s == '.adb.ald[0]'

    u, s = fetch(['x', 2, 'm'], v)
    assert u == 'j'
    assert s == '["x"][2]["m"]'

    # products
    p, r = demo_product
    chk_sample_pd(p)


def test_tree(demo_product):
    p, r = demo_product
    # test output
    ts = 'tree out_tree'
    v = tree(p)
    ts += "\n" + '\n'.join(v)
    v = tree(p, level=1)
    ts += "\n" + '\n'.join(v)
    v = tree(p, level=1, style='ascii')
    ts += "\n" + '\n'.join(v)
    if mk_outputs:
        print(ts)
        with open(output_write, 'a') as f:
            clsn = 'out_tree'
            f.write('%s = """%s"""\n' % (clsn, ts))
    else:
        assert ts == out_tree


def test_find_all_files():
    tdir = '/tmp/test_find_all_files'
    if os.path.exists(tdir) and tdir != '/':
        os.system('rm -rf '+tdir)
    os.makedirs(tdir+'/sub', exist_ok=True)
    os.system('cd %s; touch a.jsn b.json sub/c.jsn' % tdir)
    # exclude nothing
    fs = find_all_files(tdir, not_if=lambda x: False)
    assert set(fs) == {tdir+'/a.jsn', tdir+'/b.json', tdir+'/sub'}
    # exclude directories
    fs = find_all_files(tdir)
    assert set(fs) == {tdir+'/a.jsn', tdir+'/b.json'}
    # only with extension
    fs = find_all_files(tdir, include='*.jsn')
    assert set(fs) == {tdir+'/a.jsn'}
    # recursive, no directories
    fs = find_all_files(tdir, include='**/*')
    assert set(fs) == {tdir+'/a.jsn', tdir+'/b.json',
                       tdir+'/sub/c.jsn'}
    # recursive, no directories
    fs = find_all_files(tdir, include='**/*', not_if=lambda x: False)
    assert set(fs) == {tdir+'/a.jsn', tdir+'/b.json',
                       tdir+'/sub', tdir+'/sub/c.jsn'}
    # recursive, no directories, limit names
    fs = find_all_files(tdir, include='**/*.jsn')
    assert set(fs) == {tdir+'/a.jsn', tdir+'/sub/c.jsn'}
    fs = find_all_files(tdir, include='**/*.js*n')
    assert set(fs) == {tdir+'/a.jsn', tdir+'/b.json', tdir+'/sub/c.jsn'}


def test_loadcsv():
    csvf = '/tmp/fditest/testloadcsv.csv'
    a = 'as if ...'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ' ')
    assert v[0] == ('col1', ['as'], '')
    assert v[1] == ('col2', ['if'], '')
    assert v[2] == ('col3', ['...'], '')

    a = ' \t\n'+a
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ' ')
    assert v[0] == ('col1', ['as'], '')
    assert v[1] == ('col2', ['if'], '')
    assert v[2] == ('col3', ['...'], '')

    # blank line skipped
    a = a + '\n1 2. 3e3'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ' ')
    assert v[0] == ('col1', ['as', 1.0], '')
    assert v[1] == ('col2', ['if', 2.0], '')
    assert v[2] == ('col3', ['...', 3000.], '')

    # first line as header

    v = loadcsv(csvf, ' ', header=1)
    assert v[0] == ('as', [1.0], '')
    assert v[1] == ('if', [2.0], '')
    assert v[2] == ('...', [3000.], '')

    # a mixed line added. delimiter changed to ','
    a = 'as, if, ...\nm, 0.2,ev\n1, 2., 3e3'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ',', header=1)
    assert v[0] == ('as', ['m', 1.0], '')
    assert v[1] == ('if', ['0.2', 2.0], '')
    assert v[2] == ('...', ['ev', 3000.], '')

    # anothrt line added. two header lines requested -- second line taken as unit line
    a = 'as, if, ...\n A, B, R \n m, 0.2,ev\n1, 2., 3e3'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ',', header=2)
    assert v[0] == ('as', ['m', 1.0], 'A')
    assert v[1] == ('if', ['0.2', 2.0], 'B')
    assert v[2] == ('...', ['ev', 3000.], 'R')


def test_loadMedia():
    fname = 'bug.gif'
    fname = op.join(op.join(op.abspath(op.dirname(__file__)),
                            'resources'), fname)
    image = loadMedia(fname, 'image/gif')
    ho = hashlib.md5()
    ho.update(image.data)
    md5 = ho.hexdigest()
    assert md5 == '57bbbd6f8cdeafe6dc617f8969448e3b'


def test_moduleloader():

    moduleloader.main(ipath=op.abspath('tests'))


def test_fullname():
    assert fullname(Urn()) == 'fdi.pal.urn.Urn'
    assert fullname(Urn) == 'fdi.pal.urn.Urn'
    assert fullname('l') == 'str'


def test_wls():
    assert wls('') == ''
    assert wls('a') == 'a'
    assert wls('12', 2) == '12'
    assert wls('12345', 2) == '12\n34\n5'
    assert wls('12345678901234567') == '123456789012345\n67'
    # splitlines() removes trailing line-breaks
    assert wls('12345\n', 2) == '12\n34\n5'
    # no extr \n at width border
    assert wls('12\n345', 2) == '12\n34\n5'
    # no extr \n not at width border
    assert wls('1\n2345', 2) == '1\n23\n45'
    assert wls('1\n\n2345', 2) == '1\n23\n45'
    assert wls('12345\n\n', 2) == '12\n34\n5'


def obsolete_wls():
    assert wls('格式版本', 4) == '格式\n版本'
    assert wls('格式版本', 8) == '格式版本'
    assert wls('格式版本。', 10) == '格式版本。'
    assert wls('格\n式版本', 8) == '格\n式版本'
    assert wls('格式版本b', 4) == '格式\n版本\nb'
    assert wls('a格式版本', 4) == 'a格\n式版\n本'
    assert wls('格a式版本', 4) == '格a\n式版\n本'
    assert wls('格式a版本', 4) == '格式\na版\n本'
    assert wls('格式版a本', 4) == '格式\n版a\n本'
    assert wls('1\ta', 4) == '1   \n    \na'
    # unprintable activated by wide chars
    assert wls('格\ta', 4) == '格#a'
    assert wls('') == ''
    assert wls('格') == '#格'
    # \r and \r\n etc are taken out as line-breaks.
    # ref https://docs.python.org/3.6/library/stdtypes.html#str.splitlines
    # the last \n is removed
    assert wls('\r') == ''
    # page separater is out, too
    assert wls(str(b'\x1c', 'utf-8')) == ''
    assert wls('格式\x01版a本', 4) == '格式\n#版a\n本'
    # fill
    assert wls('格\n式版本', 8, fill=' ') == '格      \n式版本  '
    assert wls('格式版本b', 4, fill=' ') == '格式\n版本\nb   '
    assert wls('格\th式版本b', 5, fill=' ') == '格#h \n式版 \n本b  '
    assert wls('产品标称的起始时间', 12, fill=' ') == '产品标称的起\n始时间      '


def test_opt(caplog):
    options = [
        {'long': 'helpme', 'char': 'h', 'default': False,
         'description': 'print help'},
        {'long': 'name=', 'char': 'n', 'default': 'Boo',
         'description': 'name of ship'},
        {'long': 'verbose', 'char': 'v', 'default': True,
         'description': 'print info'}
    ]
    # no args. defaults returned
    out = opt(options, [])
    assert out[0]['result'] == False
    assert out[1]['result'] == 'Boo'
    assert out[2]['result'] == True

    assert options[1]['long'] == 'name='

    # options given in short format
    out = opt(options, ['exe', '-h', '-n Awk', '-v'])
    assert out[0]['result'] == True
    # leading and trailing white spaces in args are removed
    assert out[1]['result'] == 'Awk'
    # the switch always results in True!
    assert out[2]['result'] == True

    # options given in long format
    out = opt(options, ['exe', '--helpme', '--name=Awk', '--verbose'])
    assert out[0]['result'] == True
    assert out[1]['result'] == 'Awk'
    # the switch always results in True!
    assert out[2]['result'] == True

    # type of result is determined by that of the default
    options[0]['default'] = 0
    out = opt(options, ['exe', '--helpme', '--name=Awk', '--verbose'])
    assert out[0]['result'] == 1

    # unplanned option and '--help' get exception and exits
    with pytest.raises(getopt.GetoptError):
        out = opt(options, ['exe', '--helpme', '--name=Awk', '-y'])
    assert 'option -y not recognized' in caplog.text

    with pytest.raises(SystemExit):
        h = copy.copy(options)
        h[0]['long'] = 'help'
        out = opt(h, ['exe', '--help', '--name=Awk', '-v'])


def check_conf(cfp, typ, getConfig):
    """ plant a config file and show that it works. """

    cfn = typ + 'local.py'
    if cfp:
        cfp = op.expanduser(cfp)
        filec = op.join(cfp, cfn)
    else:
        # see test_getConfig_init
        return
    try:
        os.unlink(filec)
    except:
        pass
    for v in ('NO_PRE',):
        os.environ.pop(v, '')
    conf = 'import os; %sconfig={"jk":98, "m":os.path.abspath(__file__)}; config={"no_pre":4}' % typ
    with open(filec, 'w') as f:
        f.write(conf)
    # check conf file directory
    w = getConfig(conf=typ, force=True)
    assert w['jk'] == 98
    assert w['no_pre'] == 4
    pfile = w['m']
    assert getConfig('jk', conf=typ) == 98
    assert pfile.startswith(cfp)
    # XXX TODO: NOT WORKING
    # conf file changed afer loading
    if 0:
        try:
            pass  # os.unlink(filec)
        except:
            pass
        conf = 'import os; %sconfig={"jk":33, "m":os.path.abspath(__file__)}' % typ
        with open(filec, 'w') as f:
            f.write(conf)
        # check conf file directory
        w = getConfig(conf=typ)  # cached if w/o force
        assert w['jk'] == 98
        w = getConfig(conf=typ, force=True)  # loaded from file if force
        assert w['jk'] == 33 == getConfig('jk', conf=typ)
        os.unlink(filec)


def test_getConfig_init():

    # no arg
    v = getConfig()
    c = getConfig()
    clu = Config_Look_Up['pns']
    assert v is c
    assert all(v[k] == _v for k, _v in clu.items())
    assert all(clu[k] == _v for k, _v in v.items())

    from fdi.pns.config import pnsconfig, config as config_
    defaultconf = copy.deepcopy(pnsconfig)
    defaultconf.update(config_)
    # default settings enables default config+pnslocal
    # v is a superset
    assert all(n in v for n in defaultconf)

    # get value of normal item
    assert getConfig('scheme').startswith('http')
    assert getConfig('scheme') == clu['scheme']
    # get url
    con = getConfig()
    purl = ''.join((con['scheme'], '://',
                    con['host'], ':',
                    str(con['port']),
                    con['baseurl']
                    ))

    assert getConfig('poolurl:') == purl + '/'
    assert getConfig('poolurl:foo') == purl + '/' + 'foo'

    logging.debug('Test non-exisiting.')
    # builtin is {} means not using the default config.
    non = getConfig(conf='non-existing', builtin={}, force=True)
    assert len(non) == 0  # len(pnsconfig) + len(config_)

    f = {'foo': 9}
    # the order of the right-hand side is {foo}+pnslocal
    v_diff = dict((n, v[n]) for n in (set(v.keys()-set(defaultconf))))
    f = getConfig(builtin={'foo': 9})
    try:
        # foo is not in maps[0]
        assert f.pop('foo') == 9
    except KeyError as e:
        f['foo']
        # foo is in maps[0]
        assert f.pop('foo') == 9
    # every item in v_diff  is in f
    assert all(v == f[k] for k, v in v_diff.items())


def test_getConfig_ENV():

    # non-default conf type
    typ = 'xyz'
    # environment variable not set
    try:
        del os.environ[typ.upper()+'_CONF_DIR']
    except:
        pass

    # specify directory set a dir with ENV
    cp = '/tmp'
    # environment variable
    os.environ[typ.upper()+'_CONF_DIR'] = cp
    check_conf(cp, typ, getConfig)

    # env var overrides
    # with prefix
    before = getConfig(conf=typ)

    assert 98 == before['jk']
    assert before['username'] != 'fungi'
    # prefixed_osenv
    # now 'jk' in pnsconfig
    assert Config_Look_Up[typ].sources.maps[1]['jk'] == 98
    os.environ[typ.upper()+'_JK'] = 'shroom'
    assert typ.upper()+'_JK' in os.environ
    # with pytest.raises(KeyError):
    # a new mapping is generated without jk
    a = [len(x) for x in Config_Look_Up[typ].sources.maps]
    assert a[:4] == [1, 2, 0, 1]
    con = getConfig(conf=typ, force=True)
    b = [len(x) for x in Config_Look_Up[typ].sources.maps]
    assert b[:4] == [2, 2, 0, 1]
    assert len(Config_Look_Up[typ].cache) == 0
    assert Config_Look_Up[typ].sources.maps[0]['jk'] == 'shroom'
    assert getConfig('jk', conf=typ) == 'shroom'

    # a key not in config
    con = getConfig(conf=typ, force=True)
    assert con['username'] == 'foo'
    os.environ[typ.upper()+'_USERNAME'] = 'fungi'
    assert os.environ[typ.upper()+'_USERNAME'] == 'fungi'

    a = [len(x) for x in Config_Look_Up[typ].sources.maps]
    assert a[:4] == [2, 2, 0, 1]
    con = getConfig(conf=typ, force=True)
    b = [len(x) for x in Config_Look_Up[typ].sources.maps]
    assert b[:4] == [3, 2, 0, 1]

    # changed
    assert con['username'] == 'fungi'
    assert getConfig('username', conf=typ) == 'fungi'

    # no prefix
    assert con['no_pre'] == 4
    os.environ['NO_PRE'] = 'mollen'
    assert os.environ['NO_PRE'] == 'mollen'

    con_e = getConfig(conf=typ, force=True)
    b = [len(x) for x in Config_Look_Up[typ].sources.maps]
    assert b[:4] == [3, 2, 1, 1]
    assert con_e['no_pre'] == 'mollen'

    os.environ[typ.upper()+'_HOST'] = 'lichen'
    assert '://lichen:' in getConfig('poolurl:', conf=typ, force=True)


def test_getConfig_conf():
    # clear previous load
    importlib.invalidate_caches()
    # non-default conf type
    typ = 'abc'
    check_conf('~/.config', typ, getConfig)
    # the file has been deleted by the check_conf in the last line


def test_leapseconds():
    t0 = datetime(2019, 2, 19, 1, 2, 3, 456789, tzinfo=timezone.utc)
    assert dTAI_UTC_from_utc(t0) == timedelta(seconds=37)
    # the above just means ...
    assert utc_to_tai(t0) - t0 == timedelta(seconds=37)
    t1 = datetime(1972, 1, 1, 0, 0, 0, 000000, tzinfo=timezone.utc)
    assert dTAI_UTC_from_utc(t1) == timedelta(seconds=10)
    t2 = datetime(1970, 1, 1, 0, 0, 0, 000000, tzinfo=timezone.utc)
    # interpolation not implemented
    assert dTAI_UTC_from_utc(t2) == timedelta(seconds=4.213170)
    t3 = datetime(1968, 2, 1, 0, 0, 0, 000000, tzinfo=timezone.utc)
    assert dTAI_UTC_from_utc(t2) == timedelta(seconds=4.213170)
    t1958 = datetime(1958, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    assert dTAI_UTC_from_utc(t1958) == timedelta(seconds=0)
    # leap seconds is added on transition
    t4 = datetime(2017, 1, 1, 0, 0, 0, 000000, tzinfo=timezone.utc)
    assert utc_to_tai(t4) - utc_to_tai(t4 - timedelta(seconds=1)) == \
        timedelta(seconds=2)
    print(_fallback.cache_info())



# -*- coding: utf-8 -*-

#################
# This test is to be run on the same machine where the http pool server is running.
#################

from fdi.httppool.model.user import LOGIN_TMPLT
from fdi.dataset.unstructureddataset import UnstructuredDataset
from test_dataset import bookstore, simple_ex, complex_ex, do_jsonPath
from fdi.dataset.testproducts import get_demo_product
from fdi.dataset.serializable import serialize
from fdi.dataset.deserialize import deserialize, serialize_args
from fdi.dataset.product import Product
from fdi.pal.httpclientpool import HttpClientPool
from fdi.pal.poolmanager import PoolManager
from fdi.pal.productstorage import ProductStorage
from fdi.pal.productpool import Lock_Path_Base, makeLockpath
from fdi.utils.common import lls, trbk, fullname
from fdi.utils.fetch import fetch
from fdi.pns.fdi_requests import safe_client
from fdi.pns.jsonio import auth_headers

import pytest
import filelock

import sys
import copy
import concurrent.futures
import requests.models
from urllib.request import pathname2url
from urllib.error import URLError

from flask import current_app, g, request
from requests.auth import HTTPBasicAuth
from flask.wrappers import Response as fwResponse  # live server
# import requests
import random
import os
import requests
import pytest
from pprint import pprint
import time
from collections.abc import Mapping

from fdi.utils.options import opt


def setuplogging():
    import logging
    import logging.config
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger()

logger.setLevel(logging.INFO)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))


# last time/lastUpdate
lupd = 0


test_poolid = __name__.replace('.', '_')
prodt = 'fdi.dataset.product.Product'


if 0:
    poststr = 'curl -i -H "Content-Type: application/json" -X POST --data @%s http://localhost:5000%s --user %s'
    cmd = poststr % ('resource/' + 'nodetestinput.jsn',
                     pathname2url(pc['baseurl'] + '/' +
                                  nodetestinput['creator'] + '/' +
                                  nodetestinput['rootcause']),
                     'foo:bar')
    print(cmd)
    os.system(cmd)
    sys.exit()


def make_auth(userpass):
    return userpass
    # for mock server.
    # return
    #   "basic", {"username": userpass[0], "password": userpass[1]})


def issane(o):
    """ basic check on return """
    global lupd
    assert o is not None, "Server is having trouble"
    assert 'error' not in o, o['error']
    assert o['time'] > lupd
    lupd = o['time']


def check0result(result, msg):
    # if msg is string, an exception must have happened
    assert result == 0, 'Error %d testing script "run". msg: ' + str(msg)
    assert msg == '' or not isinstance(msg, (str, bytes)), msg


def est_getpnspoolconfig(pc, server):
    ''' gets and compares pnspoolconfig remote and local
    '''
    logger.info('get pnsconfig')
    aburl, headers = server
    o = getJsonObj(aburl + '/'+'pnsconfig')
    issane(o)
    r = o['result']
    # , deepcmp(r['scripts'], pc['scripts'])
    assert r['scripts'] == pc['scripts']
    return r


# TEST HTTPPOOL  API

def getPayload(aResponse, ignore_error=True):
    """ deserializes, if content_type is json, data or tex of responses from wither the live server or the mock one.
    """

    if not ignore_error:
        assert aResponse.status_code == 200, 'Unsuccessful response %d.' % aResponse.status_code

    x = aResponse.data if issubclass(
        aResponse.__class__, fwResponse) else aResponse.text
    if aResponse.headers['Content-Type'] == 'application/json':
        return deserialize(x, int_key=True),  aResponse.status_code
    else:
        return x, aResponse.status_code


def check_response(o, code=200, failed_case=False, excluded=None, login=False, ordered=True):
    """ Generic checking.

    :o: deserialized response data or text.
    :failed_case: True if expecing Fail result; False (default) if Success result; None if to ignore `result` being Fails or not.
    :excluded: a list of string, any of which appears in `result` is to exclude this call from being checked "FAILED".
    """
    global lupd
    if excluded is None:
        excluded = []
    assert o is not None, "Server is having trouble"
    someone = any(x in o for x in excluded)
    oc = o.__class__
    assert issubclass(oc, dict) and 'result' in o or \
        issubclass(oc, str) and 'Bad string to decode as JSON' not in o or \
        issubclass(oc, bytes) and not o.startswith(b'<!DOC') or \
        someone, o
    if failed_case is None:
        return True
    elif not failed_case:
        if not someone:
            # properly formated
            if login:
                if issubclass(oc, bytes) and o.startswith(b'<!doctype html>'):
                    assert b'password' in o
                    return True
                else:
                    return o
            assert 'FAILED' != o['result'], o['msg']
            assert code == 200, str(o)
            if ordered:
                assert o['time'] > lupd
            lupd = o['time']
            return True
    else:
        assert 'FAILED' == o['result'], o['result']
        assert code >= 400, str(o)
        return True
    return False  # not properly formated


def test_clear_local_server(local_pools_dir):

    clrpool = 'test_clear'
    ppath = os.path.join(local_pools_dir, clrpool)
    if not os.path.exists(ppath):
        os.makedirs(ppath)
    assert os.path.exists(ppath)
    with open(ppath+'/foo', 'w') as f:
        f.write('k')
    clear_server_local_pools_dir(clrpool, local_pools_dir)
    assert not os.path.exists(ppath)


def test_svcl(server, client):
    aburl, headers = server
    cl = client
    print(cl)
    print(aburl)


def test_root(server, client):
    aburl, headers = server
    url = aburl + '/'
    x = safe_client(client.get, url)
    o, code = getPayload(x)
    check_response(o, code=code)
    c0 = o['result']  # a list
    # no slash
    url = aburl
    x = safe_client(client.get, url)
    o, code = getPayload(x)  # a dict of urls
    if check_response(o, code=code, excluded=['Redirecting']):
        c = o['result']
        assert set(c0) == set(c)
    # /
    url = aburl + '/pools/'
    x = safe_client(client.get, url, headers=headers)
    o, code = getPayload(x)
    check_response(o, code=code)
    c_pools = o['result']  # a dict
    assert set(iter(c_pools)) <= set(iter(c0))


def test_wipe_all_pools_on_server(server, tmp_local_remote_pools, local_pools_dir, client, auth):

    npools = tmp_local_remote_pools
    n = len(npools)
    pool, prd, ref, tag = npools[0]
    # ======== wipe all pools =====
    logger.info('Wipe all pools on the server')

    # register all pools and get count
    aburl, headers = server
    url = aburl + '/' + 'pools/register_all'
    x = safe_client(client.put, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code, failed_case=False)
    regd = o['result']

    # make an unregistered pool by copying an existing pool
    p = os.path.join(local_pools_dir, pool.getId())
    os.system('cp -rf %s %s_copy' % (p, p))
    assert os.path.exists(p+'_copy')

    assert len(get_files_in_local_dir('', local_pools_dir)) >= n+1

    # wipe all pools
    url = aburl + '/' + 'pools/wipe_all'
    x = safe_client(client.delete, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code, failed_case=False)

    files = get_files_in_local_dir('', local_pools_dir)
    assert len(files) == 0, 'Wipe_all_pools failed: ' + \
        o['msg'] + 'Files ' + str(files)


def est_wipe_all_pools_on_server(server, local_pools_dir, client, auth):
    aburl, headers = server
    post_poolid = test_poolid
    # ======== wipe all pools =====
    logger.info('Wipe all pools on the server')

    # register all pools and get count
    url = aburl + '/' + 'pools/register_all'
    x = safe_client(client.post, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code, failed_case=False)
    regd = o['result']

    # make some pools
    n = 5
    import random
    pn = post_poolid+str(random.randint(0, 100))
    lst = make_pools(pn, aburl, client, auth, n)
    # make an unregistered pool by copying an existing pool
    p = os.path.join(local_pools_dir, lst[0].getId())
    os.system('cp -rf %s %s_copy' % (p, p))
    assert os.path.exists(p+'_copy')

    assert len(get_files_in_local_dir('', local_pools_dir)) >= n+1

    # wipe all pools
    url = aburl + '/' + 'pools/wipe_all'
    x = safe_client(client.delete, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code, failed_case=False)

    files = get_files_in_local_dir('', local_pools_dir)
    assert len(files) == 0, 'Wipe_all_pools failed: ' + \
        o['msg'] + 'Files ' + str(files)


def test_new_user_read_write(new_user_read_write, pc):
    """
    GIVEN a User model
    WHEN a new User is created
    THEN check the username, hashed_password, authenticated, and roles fields are defined correctly
    https://www.patricksoftwareblog.com/testing-a-flask-application-using-pytest/
    """
    new_user, headers = new_user_read_write
    assert new_user.username == pc['username']
    assert new_user.hashed_password != pc['password']
    assert not new_user.authenticated
    assert new_user.roles == ('read_write',)
    logger.debug('Done.')


def test_new_user_read_only(new_user_read_only, pc):
    """
    GIVEN a User model
    WHEN a new User is created
    THEN check the username, hashed_password, authenticated, and roles fields are defined correctly
    """
    new_user, headers = new_user_read_only
    assert new_user.username == pc['ro_user']
    assert not new_user.hashed_password.startswith('o')
    assert not new_user.authenticated
    assert new_user.roles == ('read_only',)
    logger.debug('Done.')


def getapis(server_ro, client):
    aburl, headers = server_ro
    x = safe_client(client.get, aburl.strip(
        '/')+'/apispec_1.json', headers=headers)
    return deserialize(x.text if type(x) == requests.models.Response else x.data)


def test_unauthorizedread_write(server, server_ro, client, tmp_local_remote_pools):
    aburl, headers = server
    roaburl, roheaders = server_ro
    pool, prod, ref, tag = tmp_local_remote_pools[0]
    poolid = pool.poolname
    # generate a unauthorized user header
    uheaders = auth_headers('k', 'hu8', headers=headers)
    if uheaders['server_type'] == 'mock':
        # we need to run before_requet, so we call full_dispatch_request
        # ref https://flask.palletsprojects.com/en/2.2.x/testing/#accessing-and-modifying-the-session
        with current_app.test_request_context(aburl+'/pools/', method='GET', headers=uheaders):
            # now we have request set
            with current_app.test_client() as client:
                # now we have client set
                with client.session_transaction() as session:
                    # now we have session set
                    user_id = session.get('user_id')
                    hdrs = str(request.headers)
                    logger.debug('*** session %x user_id = "%s"' %
                                 (id(session), str(user_id))+hdrs)
                    x = current_app.full_dispatch_request()
    else:
        x = safe_client(client.get, aburl+'/pools/', headers=uheaders)
    if LOGIN_TMPLT:
        # In order to use the login page, the return code has to be 200
        assert x.status_code == 200
        o, code = getPayload(x, ignore_error=False)
        assert str(o).startswith('<!doctype html>')
    else:
        # If json is to be used, the return code shoul be 401 to trigger browser login prompt
        assert x.status_code == 401
        o, code = getPayload(x, ignore_error=True)
        assert o in ('Unauthorized Access', b'Unauthorized Access')

    # These needs read_write
    paths = getapis(server_ro, client)['paths']
    for p, ms in paths.items():
        for meth, spec in ms.items():
            api = p.replace('{pool}', poolid)
            if meth == 'post':
                print(meth, aburl+api, '""')
                # unknown user
                x = safe_client(client.post, aburl+api,
                                headers=uheaders, data='')
                assert x.status_code == 200 if p == '/pool/{method_args}' else 401
                # read_only
                x = safe_client(client.post, roaburl+api,
                                headers=roheaders, data='')
                if LOGIN_TMPLT:
                    # In order to use the login page, the return code has to be 200
                    assert x.status_code == 200 if p == '/login' \
                        else 401 if p == '/logout' else 403
                assert x.status_code == 200 if p == '/login' \
                    else 401 if p == '/logout' else 403
                # read_write
                x = safe_client(client.post, aburl+api,
                                headers=headers, data='')
                assert x.status_code == 401 if p == '/user/logout' else 200

    logger.debug('Done.')


def test_authorizedread_write(server, new_user_read_write, client):
    aburl, headers = server
    x = safe_client(client.get, aburl+'/pools/', headers=headers)
    assert x.status_code == 200
    # with pytest.raises(URLError):
    o, code = getPayload(x)
    check_response(o, code=code, login=True)
    pools = o['result']
    assert isinstance(pools, list)


def clear_server_local_pools_dir(poolid, local_pools_dir):
    """ deletes files in the given poolid in server pool dir. """
    logger.info('clear server pool dir ' + poolid)
    path = os.path.join(local_pools_dir, poolid)
    if os.path.exists(path):
        if path == '/':
            raise ValueError('!!!!! Cannot delete root.!!!!!!!')
        else:
            os.system('rm -rf ' + path)
        # x = Product(description='desc test case')
        # x.creator = 'test'
        # data = serialize(x)
        # url = aburl + '/' + test_poolid + '/fdi.dataset.product.Product/0'
        # x = requests.post(url, auth=make_auth(userpass), data=data)


def get_files_in_local_dir(poolid, local_pools_dir):
    """ returns a list of files in the given poolid in server pool dir. """

    ppath = os.path.join(local_pools_dir, poolid)
    if os.path.exists(ppath):
        files = os.listdir(ppath)
    else:
        files = []
    return files


def empty_pool(post_poolid, aburl, auth, clnt):
    path = post_poolid + '/api/removeAll'
    url = aburl + '/' + path + '/'  # trailing / is needed by mock server
    x = clnt.get(url, auth=auth)
    o, code = getPayload(x)
    # ignore "FAILED" so non-exisiting target will not cause a failed case.
    check_response(o, code=code, failed_case=None)


def populate_pool(poolid, aburl, auth, clnt):
    creators = ['Todds', 'Cassandra', 'Jane', 'Owen', 'Julian', 'Maurice']
    instruments = ['fatman', 'herscherl', 'NASA', 'CNSC', 'SVOM']

    urns = []
    for index, i in enumerate(creators):
        x = Product(description='desc ' + str(index),
                    instrument=random.choice(instruments))
        x.creator = i
        data = serialize(x)
        url = aburl + '/' + poolid + '/'
        x = clnt.post(url, auth=auth, data=data)
        # print(len(data))
        o, code = getPayload(x)
        check_response(o, code=code)
        urns.append(o['result'])

    return creators, instruments, urns


def test_CRUD_product(local_pools_dir, server, auth, client):
    ''' test saving, read, delete products API, products will be saved at /data/pool_id
    '''

    logger.info('save products')
    aburl, headers = server

    post_poolid = test_poolid
    # auth = HTTPBasicAuth(*userpass) # not working with mock server
    # auth = tuple(userpasss.values())

    # register
    # pool = make_pools(test_poolid, aburl, client, auth, n=1)
    empty_pool(post_poolid, aburl, auth, client)

    files = [f for f in get_files_in_local_dir(
        post_poolid, local_pools_dir) if f[-1].isnumeric()]
    origin_prod = len(files)

    creators, instruments, urns = populate_pool(
        post_poolid, aburl, auth, client)

    files1 = [f for f in get_files_in_local_dir(
        post_poolid, local_pools_dir) if f[-1].isnumeric()]
    # list of would-be urns from existing files.
    urns1 = ['urn:' + post_poolid + ':' + x.replace('_', ':') for x in files1]
    num_prod = len(files1)
    assert num_prod == len(creators) + origin_prod, 'Products number not match'

    newfiles = set(files1) - set(files)
    us = set(u.split(':', 2)[2].replace(':', '_') for u in urns)
    assert newfiles == us, str(newfiles) + str(us)
    # ==========
    logger.info('read product')

    u = random.choice(urns)
    url = aburl + '/' + u  # [4:].replace(':', '/')
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    assert o['result'].creator == creators[urns.index(u)], 'Creator not match'

    # ===========
    ''' Test read hk api
    '''
    logger.info('read hk')
    hkpath = '/hk'
    url = aburl + '/' + post_poolid + hkpath + \
        '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, auth=auth)
    o1, c1 = getPayload(x)
    url2 = aburl + '/' + post_poolid + '/api/readHK' + \
        '/'  # trailing / is needed by mock server
    x2 = safe_client(client.get, url2, auth=auth)
    o2, c2 = getPayload(x2)
    for o, c in [(o1, c1), (o2, c2)]:
        check_response(o, code=c)
        assert o['result']['classes'] is not None, 'Classes jsn read failed'
        assert o['result']['tags'] is not None, 'Tags jsn read failed'
        assert o['result']['urns'] is not None, 'Urns jsn read failed'
        assert o['result']['dTypes'] is not None, 'dTypes jsn read failed'
        assert o['result']['dTags'] is not None, 'dTags jsn read failed'

        inds = list(o['result']['dTypes'][prodt]['sn'])
        l = len(inds)
        # the last l sn's
        assert list(o['result']['dTypes'][prodt]['sn'])[-l:] == inds
        assert o['result']['dTypes'][prodt]['currentSN'] == inds[-1]
        assert len(o['result']['dTags']) == 0
        assert set(':'.join(['urn', post_poolid, prodt, str(i)])
                   for i in inds) == set(urns1)

    logger.info('read dTypes')
    hkpath = '/hk/dTypes'
    url = aburl + '/' + post_poolid + hkpath
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    assert list(o['result'][prodt]['sn'])[-l:] == inds
    assert o['result'][prodt]['currentSN'] == inds[-1]
    assert set('urn:%s:%s:%s' % (post_poolid, c, str(n))
               for c in o['result'] for n in o['result'][c]['sn'].keys()) == set(urns1)

    logger.info('check count')
    num = len(o['result'][prodt]['sn'])
    apipath = '/api/getCount__' + prodt
    url = aburl + '/' + post_poolid + apipath + \
        '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    assert o['result'] == num

    logger.info('read dTags')
    hkpath = '/hk/dTags'
    url = aburl + '/' + post_poolid + hkpath
    # '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    assert len(o['result']) == 0

    # ========
    logger.info('delete a product')

    files = [f for f in get_files_in_local_dir(
        post_poolid, local_pools_dir) if f[-1].isnumeric()]
    origin_prod = len(files)

    index = files[-1].rsplit('_', 1)[1]
    # poolname following 'urn' immediately
    url = aburl + '/urn' + post_poolid + '/fdi.dataset.product.Product/' + index

    x = safe_client(client.delete, url, auth=auth)

    o, code = getPayload(x)
    check_response(o, code=code)

    files1 = [f for f in get_files_in_local_dir(
        post_poolid, local_pools_dir) if f[-1].isnumeric()]
    num_prod = len(files1)
    assert num_prod + 1 == origin_prod, 'Products number not match'

    newfiles = set(files) - set(files1)
    assert len(newfiles) == 1
    f = newfiles.pop()
    assert f.endswith(str(index))

    # ':'+poolname following 'urn'
    index2 = files[-2].rsplit('_', 1)[1]
    url2 = aburl + '/urn' + ':' + post_poolid + \
        '/fdi.dataset.product.Product/' + index2

    x = safe_client(client.delete, url2, auth=auth)

    o, code = getPayload(x)
    check_response(o, code=code)

    files2 = [f for f in get_files_in_local_dir(
        post_poolid, local_pools_dir) if f[-1].isnumeric()]
    num_prod2 = len(files2)
    assert num_prod2 + 2 == origin_prod, 'Products number not match'

    newfiles = set(files1) - set(files2)
    assert len(newfiles) == 1
    f = newfiles.pop()
    assert f.endswith(str(index2))

    # ========
    logger.info('wipe a pool')
    files = get_files_in_local_dir(post_poolid, local_pools_dir)
    assert len(files) != 0, 'Pool is already empty: ' + post_poolid

    # wipe the pool on the server
    url = aburl + '/' + post_poolid + '/api/removeAll' + \
        '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)

    files = get_files_in_local_dir(post_poolid, local_pools_dir)
    assert len(files) == 0, 'Wipe pool failed: ' + o['msg']

    url = aburl + '/' + post_poolid + '/api/isEmpty' + \
        '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    assert o['result'] == True

    # ========
    logger.info('unregister a pool on the server')
    url = aburl + '/' + post_poolid
    x = safe_client(client.delete, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)

    # this should fail as pool is unregistered on the server
    url = aburl + '/' + post_poolid + '/api/isEmpty' + \
        '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=x.status_code, failed_case=True)


def test_data_path(server, auth, client):

    aburl, headers = server
    # empty_pool(post_poolid,aburl,auth)
    pstore = ProductStorage(test_poolid, aburl + '/' +
                            test_poolid, auth=auth, client=client)
    pool = PoolManager.getPool(test_poolid)

    url0 = aburl + '/' + test_poolid + '/'
    # write sample product to the pool
    p = get_demo_product()
    prodt = fullname(p)
    data = serialize(p)
    # print(len(data))
    url1 = url0
    x = safe_client(client.post, url1, auth=auth, data=data)
    o, code = getPayload(x)
    check_response(o, code=code)
    urn = o['result']

    # API
    # url0       = 'http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/'
    # url1       = 'http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/'
    # urn        = 'urn:fdi_serv.test_httppool:fdi.dataset.product.Product:0'
    # pcls       = 'fdi.dataset.product.Product'
    # urlapi     = 'http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/fdi.dataset.product.Product'

    pcls = urn.split(':')[2].replace(':', '/')
    urlapi = url0 + pcls
    # 'http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/fdi.dataset.product.Product'
    x = safe_client(client.get, urlapi, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    c = o['result']
    assert 'metadata' in c

    # test product paths
    segs = ["measurements", "Time_Energy_Pos", "Energy", "data"]
    pth = '/'.join(segs)
    # make url w/  urn
    # url2       = 'http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/fdi.dataset.product.Product/0/measurements/Time_Energy_Pos/Energy/data'
    url2 = aburl + urn.replace(':', '/')[3:] + '/' + pth
    x = safe_client(client.get, url2, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    c = o['result']
    assert c == p['measurements']['Time_Energy_Pos']['Energy'].data
    # make w/ prodtype
    # fdi.dataset.product.Product/0
    pt = urn.split(':', 2)[2].replace(':', '/')

    urlp = url0 + pt
    # http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/fdi.dataset.product.Product/0/measurements/Time_Energy_Pos/Energy/data
    url3 = urlp + '/' + pth
    x = safe_client(client.get, url3, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    c2 = o['result']
    assert c == p['measurements']['Time_Energy_Pos']['Energy'].data

    for pth in [
            "description",
            "meta/speed/unit",
            "meta/speed/value",
            "meta/speed/isValid",
            "Environment Temperature/data",
            "measurements/calibration/unit",
    ]:
        url = urlp + '/' + pth
        x = safe_client(client.get, url, auth=auth)
        o, code = getPayload(x)
        check_response(o, code=code)
        c = o['result']
        f, s = fetch(pth, p)
        assert c == f
    # members

    # pt = fdi.dataset.product.Product/0
    url = url0 + pt + '/'
    x = safe_client(client.get, url, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code)
    c = o['result']
    assert 'description' in c

    # string

    # 'http://127.0.0.1:5000/fdi/v0.10/fdi_serv.test_httppool/string/fdi.dataset.product.Product/0'
    url = url0 + pt + '/toString'
    x = safe_client(client.get, url, auth=auth)
    assert x.headers['Content-Type'] == 'text/plain'
    o, c = getPayload(x)
    assert ('UNKNOWN' if isinstance(o, str) else b'UNKNOWN') in o


def test_get_pools(local_pools_dir, server, client):

    aburl, headers = server
    url = aburl + '/'+'pools' + '/'  # trailing / is needed by mock server
    x = safe_client(client.get, url, headers=headers)
    o, code = getPayload(x)
    check_response(o, code=code)
    c = o['result']
    assert len(c)
    assert set(c) == set(get_files_in_local_dir('', local_pools_dir))


def lock_pool2(poolid, sec, local_pools_dir):
    ''' Lock a pool from reading and return a fake response
    '''
    ppath = os.path.join(local_pools_dir, poolid)
    # lock to prevent writing
    lock = Lock_Path_Base + ppath.replace('/', '_') + '.read'
    xx = makeLockpath(poolid, 'r')
    logger.info('Keeping files %s locked for %f sec' % (lock, sec))
    t0 = str(time.time())
    with filelock.FileLock(lock):
        time.sleep(sec)
    fakeres = '{"result": "DONE", "msg": "This is a fake responses", "time": ' + t0 + '}'
    return deserialize(fakeres)


def read_product2(poolid, server, auth, client):

    aburl, headers = server

    # trying to read
    if 1:
        prodpath = '/'+prodt+'/0'
        # trailing / is needed by mock server
        url = aburl + '/' + poolid + prodpath  # + '/'
    else:
        hkpath = '/hk/dTypes'
        url = aburl + '/' + poolid + hkpath + '/'  # trailing / is needed by mock server
    logger.debug('Reading a locked file http...%s' % (url[-5:]))
    x = safe_client(client.get, url, auth=auth)
    r = x.text if type(x) == requests.models.Response else x.data
    logger.info("read %f http...%s  %s" % (time.time(), url[-5:], lls(r, 26)))
    o, code = getPayload(x)
    o['msg'] = r
    return o


def pararun(n, sleep2, *args):
    res, futs = {}, {}
    with concurrent.futures.ThreadPoolExecutor(max_workers=n) as executor:
        for i in range(n):
            time.sleep(sleep2)
            tt = time.time()
            readfut = executor.submit(*args)
            futs[tt] = readfut
        for creadfut in concurrent.futures.as_completed(futs.values()):
            # __import__('pdb').set_trace()

            readres = creadfut.result()
            tt = time.time()
            # logger.info('PPPPPP %f %s FFFF %s' % (readres['time'], readres['result'].description,st$
            # readres['msg'] = tt
            res[tt] = readres
    return res

# https://github.com/pallets/flask/issues/4375#issuecomment-990102774


def Xest_lock_file2(server, userpass, local_pools_dir, client):
    ''' Test if a pool is locked, others can not manipulate this pool anymore before it's released
    '''
    logger.info('%f Test read a locked file, it will return DONE.' %
                time.time())
    aburl, headers = server
    poolid = test_poolid
    # init server
    populate_pool(poolid, aburl, userpass, client)
    # hkpath = '/hk/dTypes'
    # url = aburl + '/' + poolid + hkpath
    # x = safe_client(client.get, url, auth=make_auth(userpass))
    locreadres, futs = {}, {}
    sleep1, sleep2 = 1.0, 0.12

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        futs[time.time()] = executor.submit(
            lock_pool2, poolid, sleep1, local_pools_dir)
        logger.info('%f lock submitted. %s. start reading ...' %
                    (time.time(), str(futs)))

        futs[time.time()] = executor.submit(pararun, 10, sleep2,
                                            read_product2, poolid,
                                            server, userpass, client)
        for locreadfut in concurrent.futures.as_completed(futs.values()):
            locreadres[time.time()] = locreadfut.result()

    logger.debug(list(locreadres))
    # __import__('pdb').set_trace()
    for t, r in locreadres.items():
        if r.get('msg', '') == 'This is a fake responses':
            # lock process
            result = r
            tl = t
        else:
            res = r
            tr = t
    logger.debug('res ' + str(res)[:99])

    # assert result['result'] == 'DONE'
    v = list(res.values())
    assert issubclass(v[0].__class__, Mapping)
    assert 'result' in v[-1]
    assert v[0]['result'].description == 'desc 0'
    assert not issubclass(v[-1]['result'].__class__, str)
    print(len(res), v[-1]['time']-result['time'])
    assert v[-1]['time']-result['time'] > sleep1
    assert len(v) > int(sleep1/sleep2) - 1

    """try:
        loop = asyncio.get_running_loop()
        # 1. Run in the default loop's executor:
        #__import__('pdb').set_trace()
        result = await loop.run_in_executor(None, lock_pool2, poolid, sleep1, local_pools_dir)
        for i in range(10):
                readres = read_product2(poolid, server, userpass, client)
                logger.info('PPPPPP %d %d' % (i,i))
                res.append(readres)
                if 'FAIL' in readres['result']:
                     time.sleep(sleep2)
                else:
                     break
        # task=functool.partial(read_product,poolid, server, userpass, client))
        # print('custom thread pool', result, res)
    except Exception as e:
        logger.error('unable to start thread ' + str(e) + trbk(e))
        raise
    """


"""
async def lock_pool(poolid, sec, local_pools_dir):
    ''' Lock a pool from reading and return a fake response
    '''
    logger.info('Keeping files locked for %f sec' % sec)
    ppath = os.path.join(local_pools_dir, poolid)
    # lock to prevent writing
    lock = Lock_Path_Base + ppath.replace('/', '_') + '.write'
    logger.debug(lock)
    with filelock.FileLock(lock):
        await asyncio.sleep(sec)
    fakeres = '{"result": "FAILED", "msg": "This is a fake responses", "time": ' + \
        str(time.time()) + '}'
    return deserialize(fakeres)


async def read_product(poolid, server, userpass, client):

    aburl, headers = server
    if headers['server_type'] == 'live':
        auth = aiohttp.BasicAuth(*userpass)
    else:
        auth = make_auth(userpass)
    # trying to read
    if 1:
        prodpath = '/'+prodt+'/0'
        # trailing / is needed by mock server
        url = aburl + '/' + poolid + prodpath + '/'
    else:
        hkpath = '/hk/dTypes'
        url = aburl + '/' + poolid + hkpath + '/'  # trailing / is needed by mock server
    logger.debug('Reading a locked file '+url)
    async with aiohttp.ClientSession() as session:
        async with session.get(url, auth=auth) as res:
            x = await res.text()
            o = deserialize(x)
    logger.debug("@@@@@@@locked file read: " + lls(x, 200))
    return o

# https://github.com/pallets/flask/issues/4375#issuecomment-990102774


def XXtest_lock_file(server, userpass, local_pools_dir, client):
    ''' Test if a pool is locked, others can not manipulate this pool anymore before it's released
    '''
    logger.info('Test read a locked file, it will return FAILED')
    aburl, headers = server
    poolid = test_poolid
    # init server
    populate_pool(poolid, aburl, userpass, client)
    # hkpath = '/hk/dTypes'
    # url = aburl + '/' + poolid + hkpath
    # x = safe_client(client.get, url, auth=make_auth(userpass))

    try:
        loop = asyncio.get_event_loop()
        tasks = [asyncio.ensure_future(
            lock_pool(poolid, 2, local_pools_dir)), asyncio.ensure_future(read_product(poolid, server, userpass, client))]
        taskres = loop.run_until_complete(asyncio.wait(tasks))
        loop.close()
    except Exception as e:
        logger.error('unable to start thread ' + str(e) + trbk(e))
        raise
    res = [f.result() for f in [x for x in taskres][0]]
    logger.debug('res ' + lls(res[0], 200) + '************' + lls(res[1], 200))
    if issubclass(res[0].__class__, Mapping) and 'result' in res[0] \
       and issubclass(res[0]['result'].__class__, str):
        r1, r2 = res[0], res[1]
    else:
        r2, r1 = res[0], res[1]
    check_response(r1, code=400, failed_case=True)
    """


def test_read_non_exists_pool(server, userpass, client):
    ''' Test read a pool which doesnot exist, returns FAILED
    '''
    logger.info('Test query a pool non exist.')
    aburl, headers = server
    wrong_poolid = 'nonexist_' + __name__.replace('.', '_')
    prodpath = '/' + prodt + '/0'
    url = aburl + '/' + wrong_poolid + prodpath
    x = safe_client(client.get, url, auth=make_auth(userpass))
    o, code = getPayload(x)
    check_response(o, code=400, failed_case=True)


def xtest_webapi_jsonPath(server, userpass, client):
    """
    """

    aburl, headers = server

    logger.info('Create pools on the server.')
    poolid = test_poolid
    poolurl = aburl + '/' + poolid
    pool = HttpClientPool(poolname=poolid, poolurl=poolurl)
    pstore = ProductStorage(pool)
    logger.info('n the server.')

    # ref
    class Get_jsonPath_from_server():
        def __init__(self, data=None, doctype='xml', attr_prefix='', *args, **kwds):
            dnm = 'bookstore' if 'bicycle' in data else 'complex_ex' if 'complex' in data else 'simple_ex'
            u = UnstructuredDataset(data=data, description=dnm,
                                    doctype=doctype, attr_prefix=attr_prefix,
                                    *args, **kwds)
            p = Product(description=dnm, data=u)
            nonlocal pool
            nonlocal pstore
            ref = pstore.save(u, tag=dnm)
            # prod  url. remove 'urn:', ':' -> '/'
            self.pool = pool
            self.purl = aburl + '/' + ref.urn[4:].replace(':', '/')

        def jsonPath(self, *args, **kwds):
            urlargs = serialize_args(*args, not_quoted=False, **kwds)
            urlargs = serialize_args(urlargs, not_quoted=False)
            url = self.purl + '/jsonPath__' + urlargs
            nonlocal userpass
            auth = make_auth(userpass)
            x = safe_client(client.get, url, auth=auth)
            o, code = getPayload(x)
            check_response(o, code=code)
            return o['result']

    do_jsonPath(Get_jsonPath_from_server)


if __name__ == '__main__':
    now = time.time()
    node, verbose = opt(pc)
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger.info('logging level %d' % (logger.getEffectiveLevel()))

    t = 8

    if t == 7:
        # test_lock()
        # asyncio.AbstractEventLoop.set_debug()
        loop = asyncio.get_event_loop()
        tasks = [asyncio.ensure_future(napa(5, 0)),
                 asyncio.ensure_future(napa(0.5, 0.5))]
        res = loop.run_until_complete(asyncio.wait(tasks))
        loop.close()
        print(res)

    elif t == 3:
        # test_getpnsconfig()
        test_puttestinit()
        test_putinit()
        test_getinit()
        test_getrun()
        test_putconfigpns()
        test_post()
        test_testrun()
        test_deleteclean()
        test_mirror()
        test_sleep()
    elif t == 4:
        test_serverinit()
        test_servertestinit()
        test_servertestrun()
        test_serversleep()
    elif t == 6:
        test_vvpp()

    print('test successful ' + str(time.time() - now))



# -*- coding: utf-8 -*-

from .metadata import Parameter
from .typecoded import Typecoded

from .finetime import FineTime, FineTime1, utcobj

from collections import OrderedDict
from copy import copy
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class DateParameter(Parameter, Typecoded):
    """ has a FineTime as the value.
    """

    def __init__(self,
                 value=None,
                 description='UNKNOWN',
                 default=None,
                 valid=None,
                 typecode=None,
                 **kwds):
        """
         Set up a parameter whose value is a point in TAI time.

        :value:
        :typecode: time format for the underlying FineTime object
        """

        # collect args-turned-local-variables.
        args = copy(locals())
        args.pop('__class__', None)
        args.pop('kwds', None)
        args.pop('self', None)
        args.update(kwds)

        # 'Q' is unsigned long long (8byte) integer.
        typecode = typecode if typecode else 'Q'
        if typecode == 'Q':
            typecode = FineTime.DEFAULT_FORMAT
        # this will set default then set value.
        super().__init__(
            value=value, description=description, typ_='finetime', default=default, valid=valid, typecode=typecode)
        # Must overwrite the self._all_attrs set by supera()
        self._all_attrs = args

    def setValue(self, value):
        """ accept any type that a FineTime does, with current typecode overriding format of the underlying FineTime
        """
        if value is not None and not issubclass(value.__class__, FineTime):
            # override format with typecode
            if hasattr(self, 'typecode'):
                tc = self.typecode
                t = FineTime(date=value, format=tc)
            else:
                t = FineTime(date=value)
        else:
            t = value
        super().setValue(t)

    def setDefault(self, default):
        """ accept any type that a FineTime does.
        """
        if default is not None and not issubclass(default.__class__, FineTime):
            default = FineTime(date=default)
        super().setDefault(default)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description if hasattr(self, 'description') else '',
                           default=self._default if hasattr(
                               self, '_default') else None,
                           value=self._value if hasattr(
                               self, '_value') else None,
                           valid=self._valid if hasattr(
                               self, '_valid') else None,
                           typecode=self.typecode if hasattr(self, 'typecode') else '')


class DateParameter1(DateParameter):
    """ Like DateParameter but usese  FineTime1. """

    def setValue(self, value):
        """ accept any type that a FineTime1 does.
        """
        if value is not None and not issubclass(value.__class__, FineTime1):
            value = FineTime1(date=value)
        super().setValue(value)

    def setDefault(self, default):
        """ accept any type that a FineTime1 does.
        """
        if default is not None and not issubclass(default.__class__, FineTime1):
            default = FineTime1(date=default)
        super().setDefault(default)



# -*- coding: utf-8 -*-

WebAPI = ('backup',
          'dereference',
          'exists',
          'getAllUrns', 'getCacheInfo', 'getCount', 'getDefinition', 'getHead',
          'getId', 'getLastVersion', 'getPlace',
          'getPoolname', 'getPoolurl', 'getPoolpath',
          'getProductClasses', 'getReferenceCount',
          'getScheme', 'getTagUrnMap', 'getTags',
          'getUrn', 'getUrnId', 'getUrnObject', 'getVersions',
          'isAlive', 'isEmpty',
          'jsonPath',
          'loadDescriptors', 'loadProduct', 'lockpath',
          'meta', 'mfilter',
          'pfilter', 'poolname', 'poolurl',
          'qm',
          'readHK', 'reference',
          'remove', 'removeAll', 'removeTag', 'removeUrn', 'removekey',
          'restore',
          'saveDescriptors', 'saveProduct', 'saveProductRef',
          'schematicLoad', 'schematicRemove', 'schematicSave', 'schematicSelect', 'schematicWipe',
          'select', 'setTag', 'setup',
          'tagExists',
          'where', 'writeHK')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# from ..pal.context import MapContext
from ..utils.options import opt
from ..utils.common import pathjoin, lls
from ..utils.ydump import yinit, ydump
from ..utils.moduleloader import SelectiveMetaFinder, installSelectiveMetaFinder
from .attributable import make_class_properties
# a dictionary that translates metadata 'type' field to classname
from .datatypes import DataTypes, DataTypeNames

# from ruamel.yaml import YAML
# import yaml
from collections import OrderedDict
import os
import sys
import functools
from itertools import chain
from string import Template
from datetime import datetime
import importlib

import logging

# create logger
logger = logging.getLogger()
logging.basicConfig(stream=sys.stdout,
                    format='%(asctime)s -%(levelname)4s'
                           ' -[%(filename)s:%(lineno)3s'
                           ' -%(funcName)10s()] - %(message)s',
                    datefmt="%Y%m%d %H:%M:%S")
logging.getLogger().setLevel(logging.DEBUG)


# make simple demo for fdi
demo = 0
# if demo is true, only output this subset.
onlyInclude = ['default', 'description',
               'data_type', 'unit', 'valid', 'fits_keyword']
# inlcude allVersions]
onlyInclude = None

# only these attributes in meta
attrs = ['startDate', 'endDate', 'instrument', 'modelName', 'mission', 'type']
indent = '    '
# extra indent
ei = ''
indents = [ei + indent * i for i in range(10)]

fmtstr = {
    'integer': '{:d}',
    'short': '{:d}',
    'hex': '0x{:02X}',
    'byte': '{:d}',
    'binary': '0b{:0b}',
    'float': '{:g}',
    'string': '"{:s}"',
    'finetime': '{:d}'
}


def sq(s):
    """ add quote mark to string, depending on if ' or " in the string.
    Parameters
    ---------

    Returns
    -------
    """

    if "'" in s or '\n' in s:
        qm = '""' if '"' in s or '\n' in s else '"'
    else:
        qm = "'"
    return '%s%s%s' % (qm, s, qm)


def get_Python(val, indents, demo, onlyInclude, debug=False):
    """ make Model and init__() code strings from given data.

    Parameters
    ----------

    Returns
    -------
    """
    infostr = ''

    if issubclass(val.__class__, dict):
        infostr += '{\n'
        code = {}
        for k, v in val.items():
            if debug:
                logger.info('KWD[%s]=%s' %
                            (str(k), '...' if k == 'metadata' else str(v)))
            sk = str(k)
            infostr += '%s%s: ' % (indents[0], sq(sk))
            if issubclass(v.__class__, dict) and 'default' in v:
                # as in k:v
                # k:
                #   data_type: string
                #   description: Description of this parent_dataset
                #   default: UNKNOWN
                #   valid: ''
                # v is a dict of parameter attributes
                istr, d_code = params(
                    v, indents[1:], demo, onlyInclude, debug=debug)
            else:
                # headers such as name, parents, schema metadata...
                istr, d_code = get_Python(
                    v, indents[1:], demo, onlyInclude, debug=debug)
            infostr += istr
            code[sk] = d_code
        infostr += indents[0] + '},\n'
    elif issubclass(val.__class__, list):
        infostr += '[\n'
        code = []
        for v in val:
            infostr += indents[0]
            if issubclass(v.__class__, dict) and 'data_type' in v:
                # val is a list of column (and 'data' in x )
                istr, d_code = params(
                    v, indents[1:], demo, onlyInclude, debug=debug)
            else:
                istr, d_code = get_Python(
                    v, indents[1:], demo, onlyInclude, debug=debug)
            infostr += istr
            code.append(d_code)
        infostr += indents[0] + '],\n'
    else:
        pval = sq(val) if issubclass(val.__class__, (str, bytes)) else str(val)
        infostr += pval + ',\n'
        code = pval
    return infostr, code


def make_init_code(dt, pval):
    """ python instanciation source code.

    will be like "default: FineTime1(0)" in 'def __init__(self...'
    Parameters
    ----------

    Returns
    -------
    """
    if dt not in ['string', 'integer', 'hex', 'binary', 'float']:
        # custom classes
        t = DataTypes[dt]
        code = '%s(%s)' % (t, pval)
    elif dt in ['integer', 'hex', 'float', 'binary']:
        code = pval
    elif pval == 'None':
        code = 'None'
    else:
        code = sq(pval)
    return code


def params(val, indents, demo, onlyInclude, debug=False):
    """ generates python strng for val, a parameter with a set of attributes

    val: as in ```name:val```
    ```
    nam:
        data_type: string
        description: Description of this parent_dataset
        default: UNKNOWN
        valid: ''
    ```
    see get_Python
    Parameters
    ----------

    Returns
    -------
    """

    # output string of the data model
    modelString = '{\n'
    # source code for init kwds.
    code = None

    # data_type
    dt = val['data_type'].strip()
    # loop through the properties
    for pname, pv in val.items():
        # pname is like 'data_type', 'default'
        # pv is like 'string', 'foo, bar, and baz', '2', '(0, 0, 0,)'
        if demo and (onlyInclude is not None) and pname not in onlyInclude:
            continue
        if debug:
            logger.info('val[%s]=%s' % (str(pname), str(pv)))
        if pname.startswith('valid'):
            if pv is None:
                pv = ''

            if issubclass(pv.__class__, (str, bytes)):
                s = sq(pv.strip())
            else:
                # e.g. {(5,66):'fooo'}
                lst = []
                for k, v in pv.items():
                    if issubclass(k.__class__, tuple):
                        fs = fmtstr[dt]
                        # (,999) in yaml is ('',999) but None from inhrited class
                        foo = [fs.format(x) if x != '' and x is not None else 'None'
                               for x in k]
                        sk = '(' + ', '.join(foo) + ')'
                    else:
                        if debug:
                            logger.info('%s: data_type %s format %s' %
                                        (pname, dt, k))
                        try:
                            sk = fmtstr[dt].format(k)
                        except TypeError:
                            sk = '# Bad format string for %s: %s. Ignored.' % (
                                dt, k)
                            logger.warning(sk)
                    lst += '\n' + '%s%s: %s,' % (indents[2], sk, sq(str(v)))
                kvs = ''.join(lst)
                if len(kvs) > 0:
                    kvs += '\n' + indents[2]
                s = '{' + kvs + '}'
        else:
            iss = issubclass(pv.__class__, (str))
            # get string representation
            pval = str(pv).strip() if iss else str(pv)
            if pname == 'default':
                code = make_init_code(dt, pval)
            if pname in ['example', 'default']:
                # here data_type instead of input type determines the output type
                iss = (val['data_type'] == 'string') and (pval != 'None')
            s = sq(pval) if iss else pval
        modelString += indents[1] + '%s: %s,\n' % (sq(pname), s)
    modelString += indents[1] + '},\n'

    return modelString, code


def get_projectclasses(clp, exclude=None, verbose=False):
    """
    return a `Classes` object that is going  to give {class-name:class-type} from a file at gieven location.

    Parameters
    ----------
    :clp: path of the mapping file.
    :rerun, exclude: from `Classes`

    Returns
    -------
    The `classes.Classes` object.
    """

    if clp is None or len(clp.strip()) == 0:
        return None

    if exclude is None:
        exclude = []
    if '/' not in clp and '\\' not in clp and not clp.endswith('.py'):
        print('Importing project classes from module '+clp)
        # classes path not given on command line
        pc = importlib.import_module(clp)
        print(
            'Imported project classes from %s module.' % clp)

    else:
        clpp, clpf = os.path.split(clp)
        sys.path.insert(0, os.path.abspath(clpp))
        # print(sys.path)
        print('Importing project classes from file '+clp)
        pc = importlib.import_module(clpf.replace('.py', ''))
        sys.path.pop(0)
    return pc


def read_yaml(ypath, version=None, verbose=False):
    """ read YAML files in ypath.

    Parameters
    ----------

    Returns
    -------
    tuple
        nm is  stem of file name. desc is descriptor, key being yaml[name]

    """
    yaml = yinit()
    desc = OrderedDict()
    fins = {}
    for findir in os.listdir(ypath):
        fin = os.path.join(ypath, findir)

        ''' The  input file name ends with '.yaml' or '.yml' (case insensitive).
        the stem name of output file is input file name stripped of the extension.
        '''
        # make it all lower case
        finl = findir.lower()
        if finl.endswith('.yml') or finl.endswith('.yaml'):
            nm = os.path.splitext(findir)[0]
        else:
            continue
        fins[nm] = fin

        # read YAML
        print('--- Reading ' + fin + '---')
        with open(fin, 'r', encoding='utf-8') as f:
            # pyYAML d = OrderedDict(yaml.load(f, Loader=yaml.FullLoader))
            ytext = f.read()
        y = yaml.load(ytext)
        d = dict(OrderedDict(y))

        if 'schema' in d and float(d['schema']) < 1.2:
            raise NotImplemented('Schema %s is too old.' % d['schema'])
        if 'metadata' not in d or d['metadata'] is None:
            d['metadata'] = {}
        if 'datasets' not in d or d['datasets'] is None:
            d['datasets'] = {}

        attrs = dict(d['metadata'])
        datasets = dict(d['datasets'])
        # move primary level table to datasets, named 'TABLE_META'
        if 'TABLE' in attrs:
            datasets['TABLE_META'] = {}
            datasets['TABLE_META']['TABLE'] = attrs['TABLE']
            del attrs['TABLE']
        if verbose:
            print('Pre-emble:\n%s' %
                  (''.join([k + '=' + str(v) + '\n'
                            for k, v in d.items() if k not in ['metadata', 'datasets']])))

        logger.debug('Find attributes:\n%s' %
                     ''.join(('%20s' % (k+'=' + str(v['default'])
                                        if 'default' in v else 'url' + ', ')
                              for k, v in attrs.items()
                              )))
        itr = ('%20s' % (k+'=' + str([c for c in (v['TABLE'] if 'TABLE'
                                                  in v else [])]))
               for k, v in datasets.items())
        logger.debug('Find datasets:\n%s' % ', '.join(itr))
        desc[d['name']] = (d, attrs, datasets, fin)
    return desc, fins


def output(nm, d, fins, version, dry_run=False, verbose=False):
    """
    Parameters
    ----------

    Returns
    -------
    """
    print("Input YAML file is to be renamed to " + fins[nm]+'.old')
    fout = fins[nm]
    print("Output YAML file is "+fout)
    if dry_run:
        print('Dry run.')
        ydump(d, sys.stdout)  # yamlfile)
    else:
        os.rename(fins[nm], fins[nm]+'.old')
        with open(fout, 'w', encoding='utf-8') as yamlfile:
            ydump(d,  yamlfile)


def yaml_upgrade(descriptors, fins, ypath, version, dry_run=False, verbose=False):
    """
    Parameters
    ----------
    :descriptors: a list of nested dicts describing the data model.
    :version: current version. not that in the yaml to be modified.
    Returns
    -------
    """

    if float(version) == 1.8:
        for nm, daf in descriptors.items():
            d, attrs, datasets, fin = daf
            if float(d['schema']) >= float(version):
                print('No need to upgrade '+d['schema'])
                continue
            d['schema'] = version
            newp = []
            for p in d['parents']:
                if p in ['Instrument', 'VT', 'VT_PDPU', 'GFT', 'GRM']:
                    newp.append('svom.instruments.' + p)
                else:
                    newp.append(p)
            d['parents'] = newp
            # increment FORMATV
            w = d['metadata']['FORMATV']
            v = w['default'].split('.')
            w.clear()
            w['default'] = version + '.' + \
                v[2] + '.' + str(int(v[3])+1)
            output(nm, d, fins, version, verbose)
    elif float(version) == 1.6:
        for nm, daf in descriptors.items():
            d, attrs, datasets, fin = daf
            if float(d['schema']) >= float(version):
                print('No need to upgrade '+d['schema'])
                continue
            d['schema'] = version
            level = d.pop('level')
            md = OrderedDict()
            for pname, w in d['metadata'].items():
                # dt = w['data_type']
                # no parent_dataset yet
                if pname == 'type':
                    v = w['default']
                    w.clear()
                    w['default'] = v
                    md[pname] = w
                    md['level'] = {'default': 'C' + level.upper()}
                elif pname == 'FORMATV':
                    v = w['default'].split('.')
                    w.clear()
                    w['default'] = version + '.' + \
                        v[2] + '.' + str(int(v[3])+1)
                    md[pname] = w
                else:
                    md[pname] = w
            d['metadata'] = md
            if 'datasets' not in d:
                d['datasets'] = {}
            output(nm, d, fins, version, dry_run=dry_run, verbose=verbose)
    else:
        logger.error('version too old')
        exit(-1)


verbo = 0


@functools.lru_cache
def mro_cmp(cn1, cn2):
    """ compare two classes by their MRO relative positions.

    Parameters
    ----------
    cn1 : str
        classname
    cn2 : str
        classname

    Returns
    -------
    int
        Returns -1 if class by the name of cn2 is a parent that of cn1,
    0 if c1 and c2 are the same class; 1 for c1 being superclasses or no relation.
    """
    if cn1 is None or cn2 is None:
        __import__('pdb').set_trace()

    if not (cn1 and cn2 and issubclass(cn1.__class__, str) and issubclass(cn2.__class__, str)):
        raise TypeError('%s and %s must be classnames.' % (str(cn1), str(cn2)))
    if verbo:
        print('...mro_cmp ', cn1, cn2)
    if cn1 == cn2:
        # quick result
        return 0
    c1 = glb[cn1]
    c2 = glb[cn2]
    if c1 is None or c2 is None:
        return None
    res = 0 if (c1 is c2) else -1 if issubclass(c1, c2) else 1
    if verbo:
        print('... ', c1, c2, res)
    return res


def descriptor_mro_cmp(nc1, nc2, des):
    """ find if nc1 is a subclasss of nc1.

    cn1 : str
        classname
    cn2 : str
        classname
    Returns
    -------
    int
        Returns -1 if class by the name of cn2 is a parent that of cn1 or its parent,
    0 if c1 and c2 are the same class; 1 for c1 being superclasses or no relation. This is to be used by `cmp_to_key` so besides having to return a negative number if `mro(nc1)` < `mro(nc2)`, it cannot return `None` for invalid situations.
  """
    if verbo:
        print('*** des_m_cmp', nc1, nc2)

    mc = mro_cmp(nc1, nc2)
    if verbo:
        print('**** mro_cmp', mc)

    if mc is None:
        if nc1 not in des:
            return 0
        # no class definition
        # 0 for top level
        parents = des[nc1][0]['parents']
        if len(parents) == 0 or len(parents) == 1 and parents[0] is None:
            return 0
        if nc1 in parents:
            raise ValueError(
                nc1 + ' cannot be its own parent. Check the YAML file.')
        if verbo:
            print(f"***** parents for {nc1} found: {parents}")
        for p in parents:
            if p == nc2:
                # This is where parent-in-des case answered.
                # parent is subclass so nc1 must be nc2's subclass
                if verbo:
                    print(f'{nc1} parent is {nc2}.')
                return -1
            else:
                dmc = descriptor_mro_cmp(p, nc2, des)
            if dmc == -1:
                # parent is subclass so nc1 must be nc2's subclass
                if verbo:
                    print(f'{nc1} parent is subc of {nc2}. d{dmc}')
                return -1
        else:
            if verbo:
                print(f'{nc1} parent is not subc of {nc2}. d{dmc}')
            return 0
    else:
        # nc1 is subclass or the same class.
        if verbo:
            print(f'{nc1} vs {nc2} => {mc}')

        return mc


def dependency_sort(descriptors):
    """ sort the descriptors so that everyone's parents are to his right.
    Parameters
    ----------

    Returns
    -------
    """
    ret = []
    # make a list of prodcts
    working_list = list(descriptors.keys())
    if verbo:
        print('+++++', str('\n'.join(working_list)))

    working_list.sort(key=functools.cmp_to_key(
        functools.partial(descriptor_mro_cmp, des=descriptors)))

    if verbo:
        print('!!!!!', str('\n'.join(working_list)))
    return working_list

    while len(working_list):
        # examin one by one
        # must use index to loop
        for i in range(len(working_list)):
            # find parents of i
            nm = working_list[i]
            # 0 for top level
            p = descriptors[nm][0]['parents']
            if nm in p:
                raise ValueError(nm + 'cannot be its own parent.')
            nm_found_parent = False
            if len(p) == 0:
                continue
            found = set(working_list) & set(p)
            # type_of_nm = glb[nm]
            # found2 = any(issubclass(type_of_nm, glb[x])
            #              for x in working_list if x != nm)
            # assert bool(len(found)) == found2
            if len(found):
                # parent is in working_list
                working_list.remove(nm)
                working_list.append(nm)
                nm_found_parent = True
                break
            else:
                # no one in the list is nm's superclass
                # TODO: only immediate parenthood tested
                ret.append(nm)
                working_list.remove(nm)
                break
            if nm_found_parent:
                break
        else:
            # no one in the list is free from deendency to others
            if len(working_list):
                msg = 'Cyclic dependency among ' + str(working_list)
                logger.error(msg)
                sys.exit(-5)
        if verbo:
            print(i, nm, p, working_list)

    return ret


def remove_Parent(a, b):
    """ Returns the one who is the other one's parent.
    Parameters
    ----------
    :a: a class name
    :b: a class name

    Returns
    -------
    classname if found or `None`.
    """
    if a == b:
        logger.debug('%s and %s are the same class' % (b, a))
        return None
    tmp = "remove parent %s because it is another parent %s's"
    if issubclass(glb[a], glb[b]):
        # remove b
        logger.debug(tmp % (b, a))
        return b
    elif issubclass(glb[b], glb[a]):
        # remove a
        logger.debug(tmp % (a, b))
        return a
    else:
        return None


def no_Parents_Parents(pn):
    """
    return a subset of class names such that no member is any other's parent.

    Parameters
    ----------
    :pn: list of class names.

    Returns
    -------
    list of non-parents.
    """

    removed = []
    for i in range(len(pn)-1):
        if pn[i] in removed:
            continue
        for j in range(i+1, len(pn)):
            r = remove_Parent(pn[i], pn[j])
            if r:
                removed.append(r)
            if r == pn[i]:
                break
    for r in removed:
        # more than one r could be in removed.
        if r in pn:
            pn.remove(r)
    return pn


def inherit_from_parents(parentNames, attrs, datasets, schema, seen):
    """ inherit metadata and datasets from parents.

    :attrs: metadata descriptor of the child
    :datasets: datasets descriptor of the child
    :seen: a dict holding class names that the py file is going to import
 """
    if parentNames and len(parentNames):
        parentsAttributes = {}
        parentsTableColumns = {}
        for parent in parentNames:
            if parent is None:
                continue
            mod_name = glb[parent].__module__
            s = 'from %s import %s' % (mod_name, parent)
            if parent not in seen:
                seen[parent] = s

            # get parent attributes and tables
            mod = sys.modules[mod_name]
            if hasattr(mod, '_Model_Spec'):
                parentsAttributes.update(
                    mod._Model_Spec['metadata'])

                parentsTableColumns.update(
                    mod._Model_Spec['datasets'])
        # merge to get all attributes including parents' and self's.
        toremove = []
        for nam, val in attrs.items():
            if float(schema) > 1.5 and 'data_type' not in val:
                # update parent's
                parentsAttributes[nam].update(attrs[nam])
                toremove.append(nam)
            else:
                # override
                parentsAttributes[nam] = attrs[nam]
        for nam in toremove:
            del attrs[nam]
        if 0 and parentsAttributes['type']['default'] == 'Oem_Nssc':
            __import__('pdb').set_trace()
        # parents are updated but names and orders follow the child's
        for ds_name, child_dataset in datasets.items():
            # go through datasets  TODO: ArrayDataset
            if ds_name not in parentsTableColumns:
                # child has a name that the parent does not have
                parentsTableColumns[ds_name] = child_dataset
                continue
            p_dset = parentsTableColumns[ds_name]
            # go through the child's dataset
            for name, c_val in child_dataset.items():
                # child has a name that the parent does not have
                if name not in p_dset:
                    p_dset[name] = c_val
                    continue
                # p and c have dataset name in common
                if name != 'TABLE':
                    # parameter in meta
                    p_dset.update(c_val)
                    continue
                p_tab = p_dset['TABLE']
                _tab = {}
                # go through the child columns
                for colname, col in c_val.items():
                    # child has a name that the parent does not have
                    if colname not in p_tab:
                        _tab[colname] = col
                        continue
                    p_tab[colname].update(col)
                    _tab[colname] = p_tab[colname]
                p_dset['TABLE'] = _tab
    else:
        parentsAttributes = attrs
        parentsTableColumns = datasets

    return parentsAttributes, parentsTableColumns


if __name__ == '__main__':

    print('Generating Python code for product class definition..')

    # schema version
    version = '1.8'

    # Get input file name etc. from command line. defaut 'Product.yml'
    cwd = os.path.abspath(os.getcwd())
    ypath = cwd
    tpath = ''
    opath = ''
    dry_run = False
    ops = [
        {'long': 'help', 'char': 'h', 'default': False, 'description': 'print help'},
        {'long': 'verbose', 'char': 'v', 'default': False,
         'description': 'print info'},
        {'long': 'yamldir=', 'char': 'y', 'default': ypath,
         'description': 'Input YAML file directory.'},
        {'long': 'template=', 'char': 't', 'default': tpath,
         'description': 'Product class template file directory. Default is the YAML dir.'},
        {'long': 'outputdir=', 'char': 'o', 'default': opath,
         'description': 'Output directory for python files. Default is the parent directory of the YAML dir.'},
        {'long': 'packagename=', 'char': 'p', 'default': '',
         'description': 'Name of the package which the generated modules belong to when imported during code generation. Default is guessing from output path.'},
        {'long': 'userclasses=', 'char': 'c', 'default': '',
         'description': 'Python file name, or a module name,  to import prjcls to update Classes with user-defined classes which YAML file refers to.'},
        {'long': 'upgrade', 'char': 'u', 'default': False,
         'description': 'Upgrade the file to current schema, by yaml_upgrade(), to version + ' + version},
        {'long': 'dry_run', 'char': 'n', 'default': False,
         'description': 'No writing. Dry run.'},
        {'long': 'debug', 'char': 'd', 'default': False,
         'description': 'run in pdb. type "c" to continuue.'},
    ]

    out = opt(ops)
    # print([(x['long'], x['result']) for x in out])
    verbose = out[1]['result']
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    ypath = out[2]['result']
    cmd_tpath = out[3]['result']
    cmd_opath = out[4]['result']
    cmd_package_name = out[5]['result']
    project_class_path = out[6]['result']
    upgrade = out[7]['result']
    dry_run = out[8]['result']
    debug = out[9]['result']

    # now can be used as parents
    from .classes import Classes

    # No bomb out if some target modules are missing.
    Classes.mapping.ignore_error = True

    # input file
    descriptors, files_imput = read_yaml(ypath, version, verbose)
    if upgrade:
        yaml_upgrade(descriptors, files_imput, ypath, version,
                     dry_run=dry_run, verbose=verbose)
        sys.exit()

    # Do not import modules that are to be generated. Thier source code
    # could be  invalid due to unseccessful previous runs
    importexclude = [x.lower() for x in descriptors.keys()]
    importinclude = {}

    # activate a module loader that refuses to load excluded
    installSelectiveMetaFinder()

    # include project classes for every product so that products made just

    pc = get_projectclasses(project_class_path,
                            exclude=importexclude, verbose=verbose)

    spupd = pc.Classes.mapping if pc else {}
    glb = Classes.update(
        c=spupd,
        exclude=importexclude,
        extension=True,
        verbose=verbose) if pc else Classes.mapping
    # make a list whose members do not depend on members behind (to the left)
    sorted_list = dependency_sort(descriptors)
    sorted_list.reverse()
    skipped = []
    for nm in sorted_list:
        d, attrs, datasets, fin = descriptors[nm]
        print('************** Processing ' + nm + '***********')

        modelName = d['name']
        # module/output file name is YAML input file "name" with lowercase
        modulename = nm.lower()

        # set paths according to each file's path
        ypath = files_imput[nm].rsplit('/', 1)[0]
        tpath = ypath if cmd_tpath == '' else cmd_tpath
        opath = os.path.abspath(os.path.join(ypath, '..')
                                ) if cmd_opath == '' else cmd_opath
        if cmd_package_name == '':
            ao = os.path.abspath(opath)
            if not ao.startswith(cwd):
                logger.error(
                    'Cannot derive package name from output dir and cwd.')
                exit(-3)
            package_name = ao[len(cwd):].strip('/').replace('/', '.')
        else:
            package_name = cmd_package_name
        logger.info("Package name: " + package_name)

        # schema
        schema = d['schema']

        # the generated source code must import these
        seen = {}
        imports = 'from collections import OrderedDict\n'
        # import parent classes
        parentNames = d['parents']
        # remove classes that are other's parent class (MRO problem)
        try:
            if parentNames and len(parentNames):
                parentNames = no_Parents_Parents(parentNames)
        except KeyError as e:
            logger.warning('!!!!!!!!!!! Skipped %s due to %s.' %
                           (nm, type(e).__name__+str(e)))
            skipped.append(nm)
            continue

        parentsAttributes, parentsTableColumns = \
            inherit_from_parents(parentNames, attrs, datasets, schema, seen)
        # make output filename, lowercase modulename + .py
        fout = pathjoin(opath, modulename + '.py')
        print("Output python file is "+fout)

        # class doc
        doc = '%s class schema %s inheriting %s.\n\nAutomatically generated from %s on %s.\n\nDescription:\n%s' % tuple(map(str, (
            modelName, schema, d['parents'],
            fin, datetime.now(), d['description'])))

        # parameter classes used in init code may need to be imported, too
        for val in chain(parentsAttributes.values(),
                         chain(col for ds in
                               parentsTableColumns.values()
                               for col in ds.get('TABLE', {}).values()
                               )
                         ):
            print(val)
            if 'data_type' not in val:

                __import__('pdb').set_trace()
            else:
                a = DataTypes[val['data_type']]
                if a in glb:
                    # this attribute class has module
                    s = 'from %s import %s' % (glb[a].__module__, a)
                    if a not in seen:
                        seen[a] = s

        # make metadata and parent_dataset dicts
        d['metadata'] = parentsAttributes
        d['datasets'] = parentsTableColumns
        infs, default_code = get_Python(d, indents[1:], demo, onlyInclude)
        # remove the ',' at the end.
        modelString = (ei + '_Model_Spec = ' + infs).strip()[:-1]

        # keyword argument for __init__
        ls = []
        for x in parentsAttributes:
            arg = 'typ_' if x == 'type' else x
            val = default_code['metadata'][x]
            ls.append(' '*17 + '%s = %s,' % (arg, val))
        ikwds = '\n'.join(ls)

        # make class properties
        properties = make_class_properties(attrs)

        # make substitution dictionary for Template
        subs = {}
        subs['WARNING'] = '# Automatically generated from %s. Do not edit.' % fin
        subs['MODELNAME'] = modelName
        print('product name: %s' % subs['MODELNAME'])
        subs['PARENTS'] = ', '.join(c for c in parentNames if c)
        print('parent class: %s' % subs['PARENTS'])
        subs['IMPORTS'] = imports + '\n'.join(seen.values())
        print('import class: %s' % ', '.join(seen.keys()))
        subs['CLASSDOC'] = doc
        subs['MODELSPEC'] = modelString
        subs['INITARGS'] = ikwds
        print('%s=\n%s\n' %
              ('Model Initialization', lls(subs['INITARGS'], 250)))
        subs['PROPERTIES'] = properties

        # subtitute the template
        if os.path.exists(os.path.join(tpath, modelName + '.template')):
            tname = os.path.join(tpath, modelName + '.template')
        elif os.path.exists(os.path.join(tpath, 'template')):
            tname = os.path.join(tpath, 'template')
        else:
            logger.error('Template file not found in %s for %s.' %
                         (tpath, modelName))
            sys.exit(-3)
        with open(tname, encoding='utf-8') as f:
            t = f.read()

        sp = Template(t).safe_substitute(subs)
        # print(sp)
        if dry_run:
            print('Dry-run. Not saving ' + fout + '\n' + '='*40)
        else:
            with open(fout, 'w', encoding='utf-8') as f:
                f.write(sp)
            print('Done saving ' + fout + '\n' + '='*40)

        # import the newly made module to test and, for class generatiom, so the following classes could use it
        importexclude.remove(modulename)
        if cwd not in sys.path:
            sys.path.insert(0, cwd)
        importlib.invalidate_caches()
        logger.debug('sys.path is '+str(sys.path))

        newp = 'fresh ' + modelName + ' from ' + modulename + \
            '.py of package ' + package_name + ' in ' + opath + '.'
        if modelName.endswith('_DataModel'):
            # the target is `Model`
            continue
        # If the last segment of package_name happens to be a module name in
        # exclude list the following import will be blocked. So lift
        # exclusion temporarily
        exclude_save = importexclude[:]
        importexclude.clear()
        try:
            _o = importlib.import_module(
                package_name + '.' + modulename, package_name)
            glb[modelName] = getattr(_o, modelName)
        except Exception as e:
            print('Unable to import ' + newp)
            raise
        importexclude.extend(exclude_save)
        print('Imported ' + newp)
        # Instantiate and dump metadata in other formats
        prod = glb[modelName]()
        # [('fancy_grid', '.txt'), ('rst', '.rst')]:
        for fmt, ext in [('fancy_grid', '.txt')]:
            fg = {'name': 15, 'value': 18, 'unit': 7, 'type': 8,
                  'valid': 26, 'default': 18, 'code': 4, 'description': 30}
            sp = prod.meta.toString(tablefmt=fmt, param_widths=fg)

            mout = pathjoin(ypath, modelName + ext)
            if dry_run:
                print('Dry-run. Not dumping ' + mout + '\n' + '*'*40)
            else:
                with open(mout, 'w', encoding='utf-8') as f:
                    f.write(sp)
                print('Done dumping ' + mout + '\n' + '*'*40)

        if len(importexclude) == 0:
            exit(0)

        Classes.update(c=importinclude, exclude=importexclude)
        glb = Classes.mapping

    if len(skipped):
        print('!!!!!!!!!!! Skipped: %s possiblly due to unresolved dependencies. Try re-running.   !!!!!!!!!!!' % str(skipped))



# -*- coding: utf-8 -*-

from serv.test_httppool import getPayload, check_response
from fdi.utils.getconfig import getConfig
from fdi.dataset.deserialize import deserialize
from fdi.dataset.product import Product
from fdi.pns.fdi_requests import safe_client, reqst, get_aio_result

import sys
import json
import os
import time
import pytest
import asyncio
import aiohttp
# from requests_threads import AsyncSession

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False


def setuplogging():
    import logging
    import logging.config
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger()

logger.setLevel(logging.INFO)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))

# https://www.twilio.com/blog/asynchronous-http-requests-in-python-with-aiohttp


@pytest.fixture
def num_pool(tmp_remote_storage_no_wipe, server, client, auth):

    ps, pool = tmp_remote_storage_no_wipe
    aburl, header = server
    poolurl = pool.poolurl
    print(pool.poolname)
    Number = 100
    return aburl, header, pool, poolurl, auth, Number


what_ = pytest.mark.skip


@what_
def xtest_cio_post(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool
    pool.removeAll()
    plist = [Product(description=str(i)).serialized() for i in range(Number)]
    urns = []

    start_time = time.time()

    async def m100():
        async with aiohttp.ClientSession() as session:
            tasks = []
            for n in range(len(plist)):
                d = plist[n]
                tasks.append(asyncio.ensure_future(
                    get_aio_result(session.post, poolurl, data=d, headers=header)))

            content = await asyncio.gather(*tasks)
            res = [deserialize(c)['result'] for code, c in content]
            print('pppp', res[0])
            print(len(res))
            return res

    urns = asyncio.run(m100())

    print("--- %s seconds ---" % (time.time() - start_time))
    assert len(urns) == Number
    res = []
    for n in range(Number):
        idx = int(urns[n].rsplit(':', 1)[1])
        print(f"{n} {idx}", end=' ')

    with open('/tmp/testurn', 'w') as f:
        json.dump(urns, f)


def test_cio_post2(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool
    pool.removeAll()
    plist = [Product(description=str(i)).serialized() for i in range(Number)]
    urns = []

    start_time = time.time()
    purls = [poolurl] * len(plist)
    urns = reqst('post', apis=purls, data=plist, headers=header)

    print("--- %s seconds ---" % (time.time() - start_time))
    assert len(urns) == Number
    res = []
    for n in range(Number):
        idx = int(urns[n].rsplit(':', 1)[1])
        print(f"{n} {idx}", end=' ')

    with open('/tmp/testurn', 'w') as f:
        json.dump(urns, f)


@what_
def xtest_cio_read(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool
    # server url
    urns = []
    try:
        with open('/tmp/testurn', 'r') as f:
            urns = json.load(f)
    except FileNotFoundError:
        urns = [""] * Number

    start_time = time.time()

    async def m100():
        async with aiohttp.ClientSession() as session:
            tasks = []
            for n in range(len(urns)):
                url = aburl+'/'+urns[n]
                tasks.append(asyncio.ensure_future(
                    get_aio_result(session.get, url, headers=header)))

            content = await asyncio.gather(*tasks)
            res = []
            for code, text in content:
                if code != 200:
                    raise RuntimeError(
                        f'AIO {method_name} error {code}: %s' % lls(text, 200))
                res.append(deserialize(text)['result'])
            print('qqqq', type(res[0]))
            print(len(res))
            return res

    res = asyncio.run(m100())
    assert len(urns) == Number
    print("--- %s seconds ---" % (time.time() - start_time))

    for n in range(Number):
        idx = int(urns[n].rsplit(':', 1)[1])
        p = res[n]
        print(f"{n} {p.description} {idx}", end=' ')


def test_cio_read2(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool
    # server url
    urns = []
    try:
        with open('/tmp/testurn', 'r') as f:
            urns = json.load(f)
    except FileNotFoundError:
        urns = [""] * Number

    start_time = time.time()

    apis = [aburl+'/'+u for u in urns]
    res = reqst('get', apis=apis, headers=header)

    assert len(urns) == Number
    print("--- %s seconds ---" % (time.time() - start_time))

    for n in range(Number):
        idx = int(urns[n].rsplit(':', 1)[1])
        p = res[n]
        print(f"{n} {p.description} {idx}", end=' ')


def test_cio_remove2(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool

    urns = pool.getAllUrns()
    print(f'remove all {len(urns)} urns in a pool with AIO.')
    start_time = time.time()

    apis = [aburl+'/'+u for u in urns]
    res = reqst('delete', apis=apis, headers=header)

    print("--- %s seconds --- %d" % (time.time() - start_time, len(res)))
    assert len(res) == Number
    assert len(pool.getAllUrns()) == 0


def est_threaded_post(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool
    pool.removeAll()
    plist = [Product(description=str(i)).serialized() for i in range(Number)]
    # server url
    refs, urns = [], []
    ppath = '/tmp/data/v0.15/test_remote_pool'
    # x = safe_client(client.get, aburl, auth=None)
    from requests_threads import AsyncSession
    session = AsyncSession(n=Number)

    async def p_x100():
        rs = []
        for n in range(Number):
            d = plist[n]
            r = await session.post(poolurl, data=d,
                                   auth=auth, headers=header)
            # urns
            urn = r.json()['result']
            # print(urn)
            rs.append(urn)
            urns.append(urn)
        print(len(rs))  # deserialize(r.content))
        # if os.path.exists(ppath):
        #     files = os.listdir(ppath)
        # else:
        #     files = []
        return rs
    # get pool

    t0 = time.time()
    if 1:
        with pytest.raises(SystemExit):
            for x in session.run(p_x100):
                continue
                print('p', x)
                urns.append(x)
    t1 = time.time()

    print('p_x100', t1-t0)
    assert len(urns) == Number
    with open('/tmp/testurn', 'w') as f:
        json.dump(urns, f)
    # refs.sorted()


def est_threaded_read(num_pool):
    aburl, header, pool, poolurl, auth, Number = num_pool
    try:
        with open('/tmp/testurn', 'r') as f:
            urns = json.load(f)
    except FileNotFoundError:
        return

    from requests_threads import AsyncSession
    session = AsyncSession(n=Number)

    async def r_x100():
        rs = []
        for n in range(len(urns)):
            r = await session.get(aburl+'/'+urns[n], auth=auth, headers=header)
            prod = deserialize(r.content)['result']
            print(n, urns[n], prod.description)
            rs.append(prod)
        return rs

    ta = time.time()
    with pytest.raises(SystemExit):
        for x in session.run(r_x100):
            pass
            prod = deserialize(x.content)
            print(prod)
    tb = time.time()
    print(tb-ta)


def est_rd100(tmp_remote_storage, server, client, auth):

    ps = tmp_remote_storage
    aburl, header = server
    pool = ps.getPool(ps.getPools()[0])
    poolurl = pool.poolurl

    Number = 10
    refs = []

    # x = safe_client(client.get, aburl, auth=None)
    session = AsyncSession(n=Number)
    # get pool

    async def r_x100():
        rs = []
        for n in range(Number):
            rs.append(await session.get(aburl, auth=None))
        print(rs)

    ta = time.time()
    with pytest.raises(SystemExit):
        for x in session.run(r_x100):
            refs.append(x)
            o, code = getPayload(x)
            # check to see if the pool url is malformed
            check_response(o, code=code, failed_case=False, ordered=False)
            # pool name is found
            assert poolname in o['reults']
        tb = time.time()
        print('@@@@ %d %.3f' % (len(refs), tb-ta))
        assert len(refs) == Number
